from __future__ import print_function
from builtins import (str, super, zip)

import numpy

from . import mixin
from . import core


class CellMethod(mixin.Properties,
                 core.CellMethod):
    '''A cell method construct of the CF data model.

One or more cell method constructs describe how the cell values of the
field construct represent the variation of the physical quantity
within its cells, i.e. the structure of the data at a higher
resolution.

A single cell method construct consists of a set of axes, a "method"
property which describes how a value of the field construct's data
array describes the variation of the quantity within a cell over those
axes (e.g. a value might represent the cell area average), and
properties serving to indicate more precisely how the method was
applied (e.g. recording the spacing of the original data, or the fact
that the method was applied only over El Nino years).

.. versionadded:: 1.7.0
    
    '''
    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

Returns a CF-netCDF-like string of the cell method.

Note that if the intention is to use this string in a CF-netCDF
cell_methods attribute then, unless they are standard names, the axes
names will need to be modified to be netCDF dimension names.

.. versionadded:: 1.7.0

        '''     
        string = ['{0}:'.format(axis) for axis in self.get_axes(())]

        string.append(self.get_method(''))

        for portion in ('within', 'where', 'over'):
            p = self.get_property(portion, None)
            if p is not None:
                string.extend((portion, p))
        #--- End: for

        interval = self.get_property('interval', ())
        comment  = self.get_property('comment', None)

        if interval:
            x = ['(']

            y = ['interval: {0}'.format(data) for data in interval]
            x.append(' '.join(y))

            if comment is not None:
                x.append(' comment: {0}'.format(comment))

            x.append(')')

            string.append(''.join(x))

        elif comment is not None:
            string.append('({0})'.format(comment))

        return ' '.join(string)
    #--- End: def

    def dump(self, display=True, _title=None, _level=0):
        '''A full description of the cell method construct.

Returns a description of all properties and the axes to which it
applies.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        indent0 = '    ' * _level

        if _title is None:
            _title = 'Cell Method: '

        return indent0 + _title + str(self)
    #--- End: def

#    def expand_intervals(self):
#        '''
#        '''
#        c = self.copy()
#
#        n_axes = len(c.get_axes(()))
#        if n_axes > 1:
#            interval = c.get_property('interval', ())            
#            if len(interval) == 1:
#                c.set_property('interval', interval*n_axes)
#
#        return c
#    #--- End: def

#    @classmethod
#    def parse(cls, string, allow_error=False):
#        '''Parse a CF cell_methods string.
#
#:Parameters:
#
#    string: `str`
#        The CF cell_methods string to be parsed into the
#        `cf.CellMethods` object. By default the cell methods will be
#        empty.
#
#    allow_error: `bool`, optional
#
#:Returns:
#
#    `list`
#
#**Examples:**
#
#>>> c = CellMethod.parse('time: minimum within years time: mean over years (ENSO years)')
#>>> print c
#Cell methods    : time: minimum within years
#                  time: mean over years (ENSO years)
#
#>>> c = CellMethod()
#>>> d = c.parse('time: minimum within years time: mean over years (ENSO years)')
#>>> print d
#Cell methods    : time: minimum within years
#                  time: mean over years (ENSO years)
#
#        '''
#        out = []
#        
#        if not string:
#            return out
#
#        # Split the cell_methods string into a list of strings ready
#        # for parsing into the result list. For example,
#        #
#        # 'lat: mean (interval: 1 hour)
#        # 
#        # maps to
#        #
#        # ['lat:', 'mean', '(', 'interval:', '1', 'hour', ')']
#        cell_methods = re_sub('\((?=[^\s])' , '( ', string)
#        cell_methods = re_sub('(?<=[^\s])\)', ' )', cell_methods).split()
#
#        while cell_methods:
#            cm = cls()
#
#            axes  = []
#            while cell_methods:
#                if not cell_methods[0].endswith(':'):
#                    break
#
#                # Check that "name" ebds with colon? How? ('lat: mean (area-weighted) or lat: mean (interval: 1 degree_north comment: area-weighted)')
#
#                axis = cell_methods.pop(0)[:-1]
#
#                axes.append(axis)
#            #--- End: while
#            cm.set_axes(axes)
#
#            if not cell_methods:
#                out.append(cm)
#                break
#
#            # Method
#            cm.set_method(cell_methods.pop(0))
#
#            if not cell_methods:
#                out.append(cm)
#                break
#
#            # Climatological statistics and statistics which apply to
#            # portions of cells
#            while cell_methods[0] in ('within', 'where', 'over'):
#                attr = cell_methods.pop(0)
#                cm.set_property(attr, cell_methods.pop(0))
#                if not cell_methods:
#                    break
#            #--- End: while
#            if not cell_methods: 
#                out.append(cm)
#                break
#
#            # interval and comment
#            interval = []
#            if cell_methods[0].endswith('('):
#                cell_methods.pop(0)
#
#                if not (re_search('^(interval|comment):$', cell_methods[0])):
#                    cell_methods.insert(0, 'comment:')
#                           
#                while not re_search('^\)$', cell_methods[0]):
#                    term = cell_methods.pop(0)[:-1]
#
#                    if term == 'interval':
#                        interval = cell_methods.pop(0)
#                        if cell_methods[0] != ')':
#                            units = cell_methods.pop(0)
#                        else:
#                            units = None
#
#                        try:
#                            parsed_interval = ast_literal_eval(interval)
#                        except:
#                            message = "Cell method interval is incorrectly formatted"
#                            if allow_error:
#                                cm = cls()
#                                cm.set_string(string)
#                                cm.set_error(message)
#                                return [out]
#                            else:
#                                raise ValueError("{}: {}".format(message, string))
#                        #---End: try
#
#                        try:
#                            interval.append(cm._Data(parsed_interval, units=units))
#                        except:
#                            message = "Cell method interval is incorrectly formatted"
#                            if allow_error:
#                                cm = cls()
#                                cm.set_string(string)
#                                cm.set_error(message)
#                                return [cm]
#                            else:
#                                raise ValueError("{}: {}".format(message, string))
#                        #---End: try
#
#                        continue
#                    #--- End: if
#
#                    if term == 'comment':
#                        comment = []
#                        while cell_methods:
#                            if cell_methods[0].endswith(')'):
#                                break
#                            if cell_methods[0].endswith(':'):
#                                break
#                            comment.append(cell_methods.pop(0))
#                        #--- End: while
#                        cm.set_property('comment', ' '.join(comment))
#                    #--- End: if
#
#                #--- End: while 
#
#                if cell_methods[0].endswith(')'):
#                    cell_methods.pop(0)
#            #--- End: if
#
#            n_intervals = len(interval)          
#            if n_intervals > 1 and n_intervals != len(axes):
#                message = "Cell method interval is incorrectly formatted"
#                if allow_error:
#                    cm = cls()
#                    cm.set_string(string)
#                    cm.set_error(message)
#                    return [out]
#                else:
#                    raise ValueError("{}: {}".format(message, string))
#            #---End: if
#
#            if interval:
#                cm.set_property('interval', tuple(interval))
#
#            out.append(cm)
#        #--- End: while
#
#        return out
#    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_properties=(), ignore_type=False):
        '''Whether two cell method constructs are the same.

Equality is strict by default. This means that:

* the descriptive properties must be the same (see the
  *ignore_properties* parameter).

The axes of the cell method constructs are *not* considered, because
they may ony be correctly interpreted by the field constructs that
contain the cell method constructs in question. They are, however,
taken into account when two fields constructs are tested for equality.

Two real numbers ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. The data type of the numbers
is not taken into consideration. See the *atol* and *rtol* parameters.

Any type of object may be tested but, in general, equality is only
possible with another cell method construct, or a subclass of one. See
the *ignore_type* parameter.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked.

.. versionadded:: 1.7.0

:Parameters:

    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_properties: sequence of `str`, optional
        The names of properties to omit from the comparison.

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another cell method construct, or a
        subclass of one. If *ignore_type* is True then
        ``CellMethod(source=other)`` is tested, rather than the
        ``other`` defined by the *other* parameter.

:Returns: 
  
    `bool`
        Whether the two cell method constructs are equal.

**Examples:**

>>> c.equals(c)
True
>>> c.equals(c.copy())
True
>>> c.equals('not a cell method')
False

        '''
        # Check that the methods are the same
        if self.get_method(None) != other.get_method(None):
            if verbose:
                print(
"{0}: Different methods: {1!r} != {2!r}".format(
    cm0.__class__.__name__, self.get_method(None), other.get_method(None)))
            return False

        ignore_properties = tuple(ignore_properties) + ('interval',)

        if not super().equals(
                other,
                rtol=rtol, atol=atol,
                verbose=verbose,
                ignore_properties=ignore_properties,
                ignore_type=ignore_type,
                ignore_data_type=True,
                ignore_fill_value=True):
            return False

#        axes0 = self.get_axes(())
#        axes1 = other.get_axes(())
#        if len(axes0) != len(axes1):
#            if verbose:
#                print(
#                    "{0}: Different axes: {1!r} != {2!r}".format(
#                        self.__class__.__name__, axes0, axes1))
#            return False
#        #--- End: if
        
        if 'interval' in ignore_properties:
            return True

        intervals0 = self.get_property('interval', ())
        intervals1 = other.get_property('interval', ())
        if intervals0:
            if not intervals1:
                if verbose:
                    print(
                        "{0}: Different intervals: {1!r} != {2!r}".format(
                            self.__class__.__name__, intervals0, intervals1))
                return False
            #--- End: if
            
            if len(intervals0) != len(intervals1):
#                intervals0 = self.expand_intervals().get_property('interval', ())
#                intervals1 = other.expand_intervals().get_property('interval', ())        
#                if len(intervals0) != len(intervals1):
                if verbose:
                    print(
                        "{0}: Different numbers of intervals: {1!r} != {2!r}".format(
                            self.__class__.__name__, intervals0, intervals1))
                    return False
            #--- End: if

            for data0, data1 in zip(intervals0, intervals1):
                if not self._equals(data0, data1,
                                    rtol=rtol, atol=atol,
                                    verbose=verbose,
                                    ignore_data_type=True,
                                    ignore_fill_value=True):
                    if verbose:
                        print(
                            "{0}: Different intervals: {1!r} != {2!r}".format(
                                self.__class__.__name__, intervals0, intervals1))
                    return False

        elif intervals1:
            if verbose:
                print("{}: Different intervals: {!r} != {!r}".format(
                    self.__class__.__name__, intervals0, intervals1))
            return False
        #--- End: if

        return True
    #--- End: def

#    def equivalent(self, other, rtol=None, atol=None, verbose=False):
#        '''True if two cell methods are equivalent, False otherwise.
#
#The `axes` and `interval` attributes are ignored in the comparison.
#
#:Parameters:
#
#    other : 
#        The object to compare for equality.
#
#    atol : float, optional
#        The absolute tolerance for all numerical comparisons, By
#        default the value returned by the `ATOL` function is used.
#
#    rtol : float, optional
#        The relative tolerance for all numerical comparisons, By
#        default the value returned by the `RTOL` function is used.
#
#:Returns: 
#
#    out : bool
#        Whether or not the two instances are equivalent.
#
#**Examples:**
#
#        '''
#        if self is other:
#            return True
#
#        # Check that each instance is the same type
#        if self.__class__ != other.__class__:
#            if verbose:
#                print("{0}: Different types: {0} != {1}".format(
#                    self.__class__.__name__, other.__class__.__name__))
#            return False
#        #--- End: if
#
#        axes0 = self.axes
#        axes1 = other.axes
#            
#        if len(axes0) != len(axes1) or set(axes0) != set(axes1):
#            if verbose:
#                print("{}: Nonequivalent axes: {!r}, {!r}".format(
#                    self.__class__.__name__, axes0, axes1))
#            return False
#        #--- End: if
#
#        argsort = [axes1.index(axis0) for axis0 in axes0]
#        other1 = other.sorted(argsort=argsort)
#        self1 = self
#
#        if not self1.equals(other1, rtol=rtol, atol=atol, ignore=('interval',)):
#            if verbose:
#                print("{0}: Nonequivalent: {1!r}, {2!r}".format(
#                    self.__class__.__name__, self, other))
#            return False
#        #--- End: if
#
#        self_interval  = self1.get_property('interval', ())
#        other_interval = other1.get_property('interval', ())
#        
#        if len(self_interval) != len(other_interval):
#            self1 = self1.expand_intervals(copy=False)
#            other1.expand_intervals(copy=False)
#
#            self_interval  = self1.get_property('interval', ())
#            other_interval = other1.get_property('interval', ())        
#
#            if len(self_interval) != len(other_interval):
#                if verbose:
#                    print(
#"{0}: Different numbers of intervals: {1!r} != {2!r}".format(
#    self.__class__.__name__, self_interval, other_interval))
#                return False
#        #--- End: if
#
##        intervals0 = self1.interval
#        if self_interval:
#            for data0, data1 in zip(self_interval, other_interval):
#                if not data0.allclose(data1, rtol=rtol, atol=atol):
#                    if verbose:
#                        print(
#"{0}: Different interval data: {1!r} != {2!r}".format(
#    self.__class__.__name__, self_interval, other_interval))
#                    return False
#        #--- End: if
#
#        # Still here? Then they are equivalent
#        return True
#    #--- End: def

    def name(self, default=None, custom=None, all_names=False):
        '''Return a name for the cell method construct.

By default the name is the first found of the following:

  1. The method, preceeded by 'method%'
  2. The value of the *default* parameter.

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        If no other name can be found then return the value of the
        default parameter. By default `None` is returned in this case.

    all_names: `bool`, optional
        If True then return a list of all possible names.

    custom: optional
        This is a dummy parameter and is ignored.
    
:Returns:

        The name. If the *all_names* parameter is True then a list of
        all possible names.

**Examples:**

>>> c
<CellMethod: domainaxis2: mean (interval: 1 day comment: ok)>
>>> c.name()
'method%mean'
>>> c.name(all_names=True, default='no name')
['method%mean', 'no name']

        '''
        out = []

        n = self.get_method(None)
        if n is not None:
            out.append('method%{0}'.format(n))
            
        if all_names:
            if default is not None:
                out.append(default)
                
            return out
        
        if out:
            return out[-1]

        return default
    #--- End: def

    def sorted(self, indices=None):
        '''Return a new cell method construct with sorted axes.
        
The axes are sorted by domain axis construct identifier or standard
name, and any intervals are sorted accordingly.

.. versionadded:: 1.7.0

:Parameters:

    indices: ordered sequence of `int`, optional
        Sort the axes with the given indices. By default the axes are
        sorted by domain axis construct identifier or standard name.

:Returns:

    `CellMethod`
        A new cell method construct with sorted axes.

**Examples:**

>>> cm = cfdm.CellMethod(axes=['domainaxis1', 'domainaxis0'],
...                      method='mean',
...                      properties={'interval': [1, 2]})
>>> cm
<CellMethod: domainaxis1: domainaxis0: mean (interval: 1 interval: 2)>
>>> cm.sorted()
<CellMethod: domainaxis0: domainaxis1: mean (interval: 2 interval: 1)>

>>> cm = cfdm.CellMethod(axes=['domainaxis0', 'area'],
...                      method='mean',
...                      properties={'interval': [1, 2]})
>>> cm
<CellMethod: domainaxis0: area: mean (interval: 1 interval: 2)>
>>> cm.sorted()
<CellMethod: area: domainaxis0: mean (interval: 2 interval: 1)>

        '''
        new = self.copy()
        
        axes = new.get_axes(())
        if len(axes) == 1:
            return new

        if indices is None:
            indices = numpy.argsort(axes)
        elif len(indices) != len(axes):
            raise ValueError(".sjdn ;siljdf vlkjndf jk")

        axes2 = []
        for i in indices:
            axes2.append(axes[i])

        new.set_axes(tuple(axes2))

        intervals = new.get_property('interval', ())
        if len(intervals) <= 1:
            return new

        intervals2 = []
        for i in indices:
            intervals2.append(intervals[i])

        new.set_property('interval', tuple(intervals2))

        return new
    #--- End: def

#    def match(self, description=None, inverse=False):
#        '''
#        '''
#        if not isinstance(description (list, tuple)):
#            description = (description,)
#
#        found_match = True
#
#        for d in description:
#            found_match = True
#            
#            if isinstance(d , basestring):
#                description = {'cell_method': d}
#                
#            c = type(self)(**d)
#
#            has_axes = False
#            if c.axes:
#                has_axes = True
#                if len(self.axes) != len(c.axes):
#                    return False
#    
#                c = c.sorted(argsort=[c.axes.index(axis) for axis in self.axes])
#                
#                if self.axes != c.axes:
#                    found_match = False
#                    continue
#            #--- End: if
#    
#            for attr in ('method', 'within', 'over', 'where', 'comment'):
#                x = getattr(c, attr)
#                if x and x != getattr(self, attr):
#                    found_match = False
#                    break
#            #--- End: for
#        
#            if not found_match:
#                continue
#    
#            if c.interval:
#                d = self.expand_intervals()
#                c.expand_intervals(copy=False)
#                    
#                intervals0 = list(self.interval)
#                intervals1 = list(c.interval)
#    
#                if len(intervals0) != len(intervals1):
#                    return False
#    
#                if has_axes:                
#                    for i0, i1 in zip(intervals0, intervals1):
#                        if i0 != i1:
#                            found_match = False
#                            break
#                else:            
#                    for i0 in intervals0:
#                        found_match = False
#                        for n, i1 in enumerate(intervals1):
#                            if i0 == i1:
#                                del intervals1[n]
#                                found_match = True
#                                break
#            #--- End: if
#
#            if not found_match:
#                continue
#        #--- End: for
#
#        return not bool(inverse)
#    #--- End: def

#--- End: class



from builtins import super

from . import mixin
from . import core


class CoordinateConversion(mixin.ParametersDomainAncillaries,
                           core.CoordinateConversion):
    '''A coordinate conversion component of a coordinate reference
construct of the CF data model.

A coordinate conversion formula converting coordinate values taken
from the dimension or auxiliary coordinate constructs to a different
coordinate system. A term of the conversion formula can be a scalar or
vector parameter which does not depend on any domain axis constructs,
may have units, or may be a descriptive string (such as the projection
name "mercator"), or it can be a reference to a domain ancillary
construct (such as one containing spatially varying orography data). A
coordinate reference construct relates the coordinate values of the
field to locations in a planetary reference frame.

.. versionadded:: 1.7.0

    '''
#--- End: class



from builtins import (str, super)

from . import abstract


class DomainAxis(abstract.Container):
    '''A domain axis construct of the CF data model. 

A domain axis construct specifies the number of points along an
independent axis of the domain. It comprises a positive integer
representing the size of the axis. In CF-netCDF it is usually defined
either by a netCDF dimension or by a scalar coordinate variable, which
implies a domain axis of size one. The field construct's data array
spans the domain axis constructs of the domain, with the optional
exception of size one axes, because their presence makes no difference
to the order of the elements.

.. versionadded:: 1.7.0

    '''
    def __init__(self, size=None, source=None, copy=True):
        '''**Initialization**

:Parameters:

    size: `int`, optional
        The size of the domain axis.

        *Parameter example:*
          ``size=192``

        The size may also be set after initialisation with the
        `set_size` method.

    source:
        Initialize the size from that of source.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__()
        
        if source is not None:
            try:
                size = source.get_size(None)
            except AttributeError:
                size = None
        #--- End: if
        
        if size is not None:
            self.set_size(size)        
    #--- End: def

    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7.0

:Returns:

    `str`
        The construct type.

        '''
        return 'domain_axis'
    #--- End: def
    
    def copy(self):
        '''Return a deep copy.

``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

.. versionadded:: 1.7.0

:Returns:	

        The deep copy.

**Examples:**

>>> e = d.copy()

        '''
        return type(self)(source=self, copy=True)
    #--- End: def

    def del_size(self, *default):
        '''Remove the size.

.. versionadded:: 1.7.0

.. seealso:: `get_size`, `has_size`, `set_size`

:Parameters:

    default: optional
        Return *default* if the size has not been set.

:Returns:

        The removed size. If unset then *default* is returned, if
        provided.

**Examples:**

>>> d.set_size(96)
>>> d.has_size()
True
>>> d.get_size()
96
>>> d.del_size()
96
>>> d.has_size()
False
>>> print(d.del_size(None))
None
>>> print(d.get_size(None))
None

        '''
        return self._del_component('size', *default)
    #--- End: def

    def has_size(self):
        '''Whether the size has been set.

.. versionadded:: 1.7.0

.. seealso:: `del_size`, `get_size`, `set_size`

:Returns:

     `bool`
        True if the size has been set, otherwise False.

**Examples:**

>>> d.set_size(96)
>>> d.has_size()
True
>>> d.get_size()
96
>>> d.del_size()
96
>>> d.has_size()
False
>>> print(d.del_size(None))
None
>>> print(d.get_size(None))
None

        '''
        return self._has_component('size')
    #--- End: def

    def get_size(self, *default):
        '''Return the size.

.. versionadded:: 1.7.0

.. seealso:: `del_size`, `has_size`, `set_size`

:Parameters:

    default: optional
        Return *default* if the size has not been set.

:Returns:

        The size. If unset then *default* is returned, if provided.

**Examples:**

>>> d.set_size(96)
>>> d.has_size()
True
>>> d.get_size()
96
>>> d.del_size()
96
>>> d.has_size()
False
>>> print(d.del_size(None))
None
>>> print(d.get_size(None))
None

        '''
        return self._get_component('size', *default)
    #--- End: def

    def set_size(self, size, copy=True):
        '''Set the size.

.. versionadded:: 1.7.0

.. seealso:: `del_size`, `get_size`, `has_size`

:Parameters:

    value: `int`
        The size.

    copy: `bool`, optional
        If True then set a deep copy of *size*.

:Returns:

     `None`

**Examples:**

>>> d.set_size(96)
>>> d.has_size()
True
>>> d.get_size()
96
>>> d.del_size()
96
>>> d.has_size()
False
>>> print(d.del_size(None))
None
>>> print(d.get_size(None))

        '''
        self._set_component('size', size, copy=copy)
    #--- End: def

#--- End: class



from builtins import super

import numpy

from . import abstract


class NumpyArray(abstract.Array):
    '''A container for a numpy array.

.. versionadded:: 1.7.0

    '''
    def __init__(self, array=None):
        '''**Initialization**

:Parameters:

    array: `numpy.ndarray`
        The numpy array.

        '''
        super().__init__(array=array)
    #--- End: def
    
    @property
    def dtype(self):
        '''Data-type of the data elements.

.. versionadded:: 1.7.0

**Examples:**

>>> a.dtype
dtype('float64')
>>> print(type(a.dtype))
<type 'numpy.dtype'>

        '''
#        return self.array.dtype
        return self._get_component('array').dtype
    #--- End: def

    @property
    def ndim(self):
        '''Number of array dimensions
                
.. versionadded:: 1.7.0

**Examples:**

>>> a.shape
(73, 96)
>>> a.ndim
2
>>> a.size
7008

>>> a.shape
(1, 1, 1)
>>> a.ndim
3
>>> a.size
1

>>> a.shape
()
>>> a.ndim
0
>>> a.size
1

        '''
        return self._get_component('array').ndim
    #--- End: def
    
    @property
    def shape(self):
        '''Tuple of array dimension sizes.

.. versionadded:: 1.7.0

**Examples:**

>>> a.shape
(73, 96)
>>> a.ndim
2
>>> a.size
7008

>>> a.shape
(1, 1, 1)
>>> a.ndim
3
>>> a.size
1

>>> a.shape
()
>>> a.ndim
0
>>> a.size
1

'''
        return self._get_component('array').shape
    #--- End: def
    
    @property
    def size(self):
        '''Number of elements in the array.

.. versionadded:: 1.7.0

**Examples:**

>>> a.shape
(73, 96)
>>> a.size
7008
>>> a.ndim
2

>>> a.shape
(1, 1, 1)
>>> a.ndim
3
>>> a.size
1

>>> a.shape
()
>>> a.ndim
0
>>> a.size
1
        '''
        return self._get_component('array').size
    #--- End: def
    
    def get_array(self):
        '''Return an independent numpy array containing the data.

.. versionadded:: 1.7.0

:Returns:

    `numpy.ndarray`
        An independent numpy array of the data.

**Examples:**

>>> n = numpy.asanyarray(a)
>>> isinstance(n, numpy.ndarray)
True

        '''
        array = self._get_component('array')
        
        if not array.ndim and numpy.ma.isMA(array):
            # This is because numpy.ma.copy doesn't work for
            # scalar arrays (at the moment, at least)
            ma_array = numpy.ma.empty((), dtype=array.dtype)
            ma_array[...] = array
            array = ma_array
        else:
            array = array.copy()

        return array
    #--- End: def

#--- End: class



from __future__ import print_function
from builtins import (super, zip)
from past.builtins import basestring

from . import core


class Constructs(core.Constructs):
    '''TODO

.. versionadded:: 1.7.0

    '''    
    def constructs(self, name=None, properties=None, measure=None,
                   ncvar=None, ncdim=None, key=None, axis=None,
                   construct_type=None, copy=False):
        '''Return metadata constructs

By default all metadata constructs are returned, but a subset may be
selected via the optional parameters. If multiple parameters are
specified, then the constructs that satisfy *all* of the criteria are
returned. 

.. versionadded:: 1.7.0

.. seealso:: `del_construct`, `get_construct`,
             `get_construct_data_axes`, `get_construct_id`,
             `has_construct`, `set_construct`

:Parameters:

    name: (sequence of) `str`, optional
        Select constructs that have the given name. If a sequence of
        names has been given then the constructs that have any of the
        names are selected.
       
        In general, a construct is selected if any of the given names
        is the same as one of the possible constuct names, as returned
        by the construct's `!name` method with the ``all_names``
        parameter set to True. For example, the following construct,
        ``c``, has three default names:

           >>> c.name(all_names=True)
           ['longitude', 'long_name:Longitude', 'ncvar%lon']

        Note that the names used to identify metadata constructs in
        the ouput of a `print` or `!dump` call may always be used to
        select constructs.

        A name may be one of:

        * The value of the standard name property.

          *Example:*
            ``name='air_pressure'`` will select constructs that have a
            "standard_name" property with the value "air_pressure".

        * The value of any property prefixed by the property name and
          a colon (``:``).

          *Example:*
            ``name='long_name:Air Temperature'`` will select
            constructs that have a "long_name" property with the value
            "Air Temperature".

          *Example:*
            ``name='positive:up'`` will select constructs that have a
            "positive" property with the value "up".

          *Example:*
            ``name='foo:bar'`` will select constructs that have a
            "foo" property with the value "bar".

          *Example:*
            ``name='standard_name:air_pressure'`` will select
            constructs that have a "standard_name" property with the
            value "air_pressure". Note this selection could also be
            made with ``name='air_pressure'``.

        * The measure of cell measure constructs, prefixed by
          ``measure%``. Constructs may also be selected by their
          meaure with the *measure* parameter.

          *Example:*
            ``name='measure%area'`` will select "area" cell measure
            constructs. Note this selection could also be made with
            ``measure='area'``.


        * A construct identifier, prefixed by ``key%`` (see also the
          *key* parameter). Constructs may also be selected by their
          *construct identifier with the *key* parameter.


          *Example:* 
            ``name='key%cellmethod1'`` will select cell method
            construct with construct identifier "cellmethod1".

        * The netCDF variable name, prefixed by ``ncvar%``.
          Constructs may also be selected by their netCDF variable
          name with the *ncvar* parameter.

          *Example:*
            ``name='ncvar%lat'`` will select constructs with netCDF
            variable name "lat".

        * The netCDF dimension name of domain axis constructs,
          prefixed by ``ncdim%``. Constructs may also be selected by
          their netCDF dimension name with the *ncdim* parameter.

          *Example:*
            ``name='ncdim%time'`` will select domain axis constructs
            with netCDF dimension name "time".

    construct_type: (sequence of) `str`, optional
        Select constructs of the given type. If a sequence of types
        has been given then the constructs of each type are
        selected. Valid types are:

          ==========================  ================================
          *construct_type*            Selected constructs
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        *Example:*
          ``construct_type='dimension_coordinate'``

        *Example:*
          ``construct_type=['auxiliary_coordinate']``

        *Example:*
          ``construct_type=('domain_ancillary', 'cell_method')``

        Note that a domain never contains cell method nor field
        ancillary constructs.

    measure: (sequence of) `str`, optional
        Select cell measure constructs which have the given
        measure. If multiple measures are specified then select the
        cell measure constructs which have any of the given measures.

        *Example:*
          ``meausure='area'``

        *Example:*
          ``measure=['area']``

        *Example:*
          ``measure=['area', 'volume']``

    axis: (sequence of) `str`, optional
        Select constructs which have data that spans a domain axis
        construct, defined by its construct identifier. If multiple of
        domain axes are specified then select constructs whose data
        spans at least one the domain axis constructs.

        *Example:*
          ``axis='domainaxis1'``

        *Example:*
          ``axis=['domainaxis2']``

        *Example:*
          ``axis=['domainaxis0', 'domainaxis1']``

    ncvar: (sequence of) `str`, optional
        Select constructs which have the given netCDF variable
        name. If multiple netCDF variable names are specified then
        select the constructs which have any of the given netCDF
        variable names.

        *Example:*
          ``ncvar='lon'``

        *Example:*
          ``ncvar=['lat']``

        *Example:*
          ``ncvar=['lon', 'lat']``

    ncdim: (sequence of) `str`, optional
        Select domain axis constructs which have the given netCDF
        dimension name. If multiple netCDF dimension names are
        specified then select the domain axis constructs which have
        any of the given netCDF dimension names.

        *Example:*
          ``ncdim='lon'``

        *Example:*
          ``ncdim=['lat']``

        *Example:*
          ``ncdim=['lon', 'lat']``


    key: (sequence of) `str`, optional
        Select the construct with the given construct key. If multiple
        keys are specified then select all of the metadata constructs
        which have any of the given keys.

        *Example:*
          ``key='domainancillary0'``

        *Example:*
          ``key=['cellmethod2']``

        *Example:*
          ``key=('dimensioncoordinate1', 'fieldancillary0')``

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    out: `dict`
        Constructs are returned as values of a dictionary, keyed by
        their construct identifiers.
        
        If cell method contructs, and no other construct types, have
        been selected with the *construct_type* parameter then the
        constructs are returned in an ordered dictionary
        (`collections.OrderedDict`). The order is determined by the
        order in which the cell method constructs were originally
        added.

**Examples:**

>>> f.constructs()
{}

>>> f.constructs()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >,
 'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>,
 'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>,
 'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>,
 'domainaxis0': <DomainAxis: 1>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>,
 'domainaxis3': <DomainAxis: 1>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs('grid_latitude')
{'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>}
>>> f.constructs('long_name:Grid latitude name')
{'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >}
>>> f.constructs('ncvar%b')
{'domainancillary1': <DomainAncillary: ncvar%b(1) >}
>>> f.constructs(construct_type='coordinate_reference')
{'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>}
>>> f.constructs(construct_type='cell_method')
OrderedDict([('cellmethod0', <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>),
             ('cellmethod1', <CellMethod: domainaxis3: maximum>)])
>>> f.constructs(construct_type=['cell_method', 'field_ancillary'])
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs(axes=['domainaxis0'])
{'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >}
>>> f.constructs(axes=['domainaxis0', 'domainaxis1'])
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >,
 'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>,
 'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}
>>> f.constructs('longitude',
...              construct_type='auxiliary_coordinate', 
...              axes=['domainaxis1'])
{'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>}
>>> f.constructs('air_pressure')
{}

        '''
        out = super().constructs(construct_type=construct_type, copy=copy)

        if key is not None:
            if isinstance(key, basestring):
                key = (key,)

            out = {cid: out[cid] for cid in key if cid in out}
            if not out:
                return out
        #--- End: if

        if axis is not None:
            if isinstance(axis, basestring):
                axis = (axis,)

            axis = set(axis)
            
            constructs_data_axes = self.constructs_data_axes()
            for cid in tuple(out):
                x = constructs_data_axes.get(cid)
                if x is None or not axis.intersection(x):
                    # This construct does not span these axes
                    del out[cid]
            #--- End: for
                        
            if not out:
                return out
        #--- End: if

        if name is not None:
            if isinstance(name, basestring):
                name = (name,)

            for cid, construct in tuple(out.items()):
                ok = False                
                for n in name:
                    (prefix, _, value) = n.partition('%')
                    if prefix == 'key':
                        if value == cid:
                            # This construct matches this name
                            ok = True
                            break
                    else:
                        (prefix, _, value) = n.partition(':')
                        custom = (prefix,) if value else None
                        if n in construct.name(custom=custom,
                                               all_names=True):
                            # This construct matches this name
                            ok = True
                            break
                #--- End: for
                
                if not ok:
                    # This construct does not match any of the names
                    del out[cid]
            #--- End: for

            if not out:
                return out
        #--- End: if

        if properties is not None:
            if isinstance(properties, dict):
                properties = (properties,)

            for cid, construct in tuple(out.items()):
                try:
                    get_property = construct.get_property
                except AttributeError:
                    del out[cid]
                    continue
                
                ok = False
                for props in properties:
                    ok = False
                    for p, value0 in props.items():
                        value1 = get_property(p, None)
                        if value1 is None or not construct._equals(value1, value0):
                            # This construct does not match this set
                            # of properties
                            ok = False                            
                            break
                        else:
                            ok = True
                    #--- End: for

                    if ok:
                        # This construct matches this set of properties
                        break
                #--- End: for
                
                if not ok:
                    # This construct does not match any of the sets of
                    # properties
                    del out[cid]
            #--- End: for
            
            if not out:
                return out
        #--- End: if

        if measure is not None:
            if isinstance(measure , basestring):
                measure = (measure,)

            for cid, construct in tuple(out.items()):
                try:
                    get_measure = construct.get_measure
                except AttributeError:
                    del out[cid]
                    continue

                ok = False
                for x0 in measure:
                    x1 = get_measure(None)
                    if x1 is not None and construct._equals(x1, x0):
                        # This construct matches this measure
                        ok = True
                        break
                #--- End: for
                
                if not ok:
                    # This construct does not match any of the
                    # measures
                    del out[cid]
            #--- End: for
            
            if not out:
                return out
        #--- End: if
        
        if ncvar is not None:
            if isinstance(ncvar , basestring):
                ncvar = (ncvar,)

            for cid, construct in tuple(out.items()):
                try:
                    nc_get_variable = construct.nc_get_variable
                except AttributeError:
                    del out[cid]
                    continue

                ok = False
                for x0 in ncvar:
                    x1 = nc_get_variable(None)
                    if x1 is not None and construct._equals(x1, x0):
                        # This construct matches this netCDF variable
                        # name
                        ok = True
                        break
                #--- End: for
                
                if not ok:
                    # This construct does not match any of the netCDF
                    # variable names
                    del out[cid]
            #--- End: for
            
            if not out:
                return out
        #--- End: if
                
        if ncdim is not None:
            if isinstance(ncdim , basestring):
                ncdim = (ncdim,)
                
            for cid, construct in tuple(out.items()):
                try:
                    nc_get_dimension = construct.nc_get_dimension
                except AttributeError:
                    del out[cid]
                    continue

                ok = False
                for x0 in ncdim:
                    x1 = nc_get_dimension(None)
                    if x1 is not None and construct._equals(x1, x0):
                        # This construct matches this netCDF dimension
                        # name
                        ok = True
                        break
                #--- End: for
                
                if not ok:
                    # This construct does not match any of the netCDF
                    # dimension names
                    del out[cid]
            #--- End: for
            
            if not out:
                return out
        #--- End: if
        
        return out
    #--- End: def

    def domain_axis_name(self, axis):
        '''Return the canonical name for an axis.

:Parameters:

    axis: `str`
        The identifier of the axis.

          *Example:*
            ``axis='domainaxis2'``

:Returns:

    out: `str`
        The canonical name for the axis.

**Examples:**

>>> f.domain_axis_name('domainaxis1')
'longitude'


        '''
        domain_axes = self.constructs(construct_type='domain_axis')
        
        if axis not in domain_axes:
            return default

        constructs_data_axes = self.constructs_data_axes()

        dimension_coordinates = self.constructs(construct_type='dimension_coordinate')

        name = None        
        for key, dim in dimension_coordinates.items():
            if constructs_data_axes[key] == (axis,):
                # Get the name from a dimension coordinate
                name = dim.name(ncvar=False, default=None)
                break
        #--- End: for
        if name is not None:
            return name

        auxiliary_coordinates = self.constructs(construct_type='auxiliary_coordinate')
        
        found = False
        for key, aux in auxiliary_coordinates.items():
            if constructs_data_axes[key] == (axis,):
                if found:
                    name = None
                    break
                
                # Get the name from an auxiliary coordinate
                name = aux.name(ncvar=False, default=None)
                found = True
        #--- End: for
        if name is not None:
            return name

        ncdim = domain_axes[axis].nc_get_dimension(None)
        if ncdim is not None:
            # Get the name from a netCDF dimension
            return 'ncdim%{0}'.format(ncdim)

        # Get the name from the identifier
        return 'key%{0}'.format(axis)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_compression=False, ignore_type=False):
        '''TODO

.. versionadded:: 1.7.0
        
        '''
        if self is other:
            return True
        
        # Check that each instance is the same type
        if type(self) != type(other):
            if verbose:
                print("{0}: Different object types: {0}, {1}".format(
                    self.__class__.__name__, other.__class__.__name__))
            return False
        
        axes0_to_axes1 = {}
        axis0_to_axis1 = {}
        axis1_to_axis0 = {}
        key1_to_key0   = {}
        
        # ------------------------------------------------------------
        # Domain axis constructs
        # ------------------------------------------------------------
        if not self._equals_domain_axis(other, rtol=rtol, atol=atol,
                                        verbose=verbose,
                                        ignore_type=ignore_type,
                                        axis1_to_axis0=axis1_to_axis0,
                                        key1_to_key0=key1_to_key0):
            return False
        
        # ------------------------------------------------------------
        # Constructs with arrays
        # ------------------------------------------------------------
        axes_to_constructs0 = self.axes_to_constructs()
        axes_to_constructs1 = other.axes_to_constructs()
        for axes0, constructs0 in axes_to_constructs0.items():
            matched_all_constructs_with_these_axes = False

            log = []
            
            len_axes0 = len(axes0) 
            for axes1, constructs1 in tuple(axes_to_constructs1.items()):
                log = []
                constructs1 = constructs1.copy()

                matched_roles = False

                if len_axes0 != len(axes1):
                    # axes1 and axes0 contain differents number of
                    # domain axes.
                    continue

                for construct_type in self._array_constructs:
                    matched_role = False
                    role_constructs0 = constructs0[construct_type]
                    role_constructs1 = constructs1[construct_type].copy()

                    if len(role_constructs0) != len(role_constructs1):
                        # There are the different numbers of
                        # constructs of this type
                        matched_all_constructs_with_these_axes = False
                        log.append('Different numbers of '+construct_type)
                        break

                    # Check that there are matching pairs of equal
                    # constructs
                    for key0, item0 in role_constructs0.items():
                        matched_construct = False
                        for key1, item1 in tuple(role_constructs1.items()):
                            if item0.equals(item1,
                                            rtol=rtol, atol=atol,
                                            verbose=False,
                                            ignore_data_type=ignore_data_type,
                                            ignore_fill_value=ignore_fill_value,
                                            ignore_compression=ignore_compression,
                                            ignore_type=ignore_type):
                                del role_constructs1[key1]
                                key1_to_key0[key1] = key0
                                matched_construct = True
                                break
                        #--- End: for

                        if not matched_construct:
                            log.append("Can't match "+construct_type+" "+repr(item0))
                            break
                    #--- End: for

                    if role_constructs1:
                        # At least one construct in other is not equal
                        # to a construct in self
                        break

                    # Still here? Then all constructs of this type
                    # that spanning these axes match
                    del constructs1[construct_type]
                #--- End: for

                matched_all_constructs_with_these_axes = not constructs1

                if matched_all_constructs_with_these_axes:
                    del axes_to_constructs1[axes1]
                    break
            #--- End: for

            if not matched_all_constructs_with_these_axes:
                if verbose:
                    names = [self.domain_axis_name(axis0) for axis0 in axes0]
                    print("Can't match constructs spanning axes {0}".format(names))
                    print('\n'.join(log))
                    print()
                    print(axes_to_constructs0)
                    print()
                    print(axes_to_constructs1)
                return False

            # Map item axes in the two instances
            axes0_to_axes1[axes0] = axes1
        #--- End: for

        for axes0, axes1 in axes0_to_axes1.items():
            for axis0, axis1 in zip(axes0, axes1):
                if axis0 in axis0_to_axis1 and axis1 != axis0_to_axis1[axis0]:
                    if verbose:
                        print(
"Field: Ambiguous axis mapping ({} -> both {} and {})".format(
    self.domain_axis_name(axes0), other.domain_axis_name(axis1),
    other.domain_axis_name(axis0_to_axis1[axis0])))
                    return False
                elif axis1 in axis1_to_axis0 and axis0 != axis1_to_axis0[axis1]:
                    if verbose:
                        print(
"Field: Ambiguous axis mapping ({} -> both {} and {})".format(
    self.domain_axis_name(axis0), self.domain_axis_name(axis1_to_axis0[axis0]),
    other.domain_axis_name(axes1)))
                    return False

                axis0_to_axis1[axis0] = axis1
                axis1_to_axis0[axis1] = axis0
        #--- End: for

        # ------------------------------------------------------------
        # Constructs with no arrays
        # ------------------------------------------------------------
        for construct_type in self._non_array_constructs:
            if not getattr(self, '_equals_'+construct_type)(
                    other,
                    rtol=rtol, atol=atol,
                    verbose=verbose,
                    ignore_type=ignore_type,
                    axis1_to_axis0=axis1_to_axis0,
                    key1_to_key0=key1_to_key0):
                return False

        # ------------------------------------------------------------
        # Still here? Then the two objects are equal
        # ------------------------------------------------------------     
        return True
    #--- End: def
    
    def _equals_coordinate_reference(self, other, rtol=None, atol=None,
                                     verbose=False,
                                     ignore_type=False,
                                     axis1_to_axis0=None,
                                     key1_to_key0=None):
        '''
        '''
        refs0 = self.constructs(construct_type='coordinate_reference')
        refs1 = other.constructs(construct_type='coordinate_reference')

        if len(refs0) != len(refs1):
            if verbose:
                print(
"Verbose: Different coordinate references: {0!r}, {1!r}".format(
    list(refs0.values()), list(refs1.values())))
            return False

        if refs0:
            for ref0 in refs0.values():
                found_match = False
                for key1, ref1 in tuple(refs1.items()):
                    if not ref0.equals(ref1, rtol=rtol, atol=atol,
                                       verbose=False,
                                       ignore_type=ignore_type):
                        continue

                    # Coordinates
                    coordinates0 = ref0.coordinates()
                    coordinates1 = set()
                    for value in ref1.coordinates():
                        coordinates1.add(key1_to_key0.get(value, value))
                        
                    if coordinates0 != coordinates1:
                        continue
    
                    # Domain ancillary-valued coordinate conversion terms
                    terms0 = ref0.coordinate_conversion.domain_ancillaries()
#                    terms1 = {}
#                    for term, key in tuple(ref1.coordinate_conversion.domain_ancillaries().items()):
#                        terms1[term] = key1_to_key0.get(key, key)

                    terms1 = {term: key1_to_key0.get(key, key)
                              for term, key in ref1.coordinate_conversion.domain_ancillaries().items()}
#                    print(terms1)

                    if terms0 != terms1:
                        continue
    
#                    # Domain ancillary-valued datum terms
#                    terms0 = ref0.datum.domain_ancillaries()
#                    terms1 = {}
#                    for term, key in ref1.datum.domain_ancillaries().items():
#                        terms1[term] = key1_to_key0.get(key, key)
#    
#                    if terms0 != terms1:
#                        continue
    
                    found_match = True
                    del refs1[key1]                                       
                    break
                #--- End: for
    
                if not found_match:
                    if verbose:
                        print(
"Verbose: No match for {0!r})".format(ref0))
                    return False
            #--- End: for
        #--- End: if

        return True
    #--- End: def

    def _equals_cell_method(self, other, rtol=None, atol=None,
                            verbose=False, ignore_type=False,
                            axis1_to_axis0=None, key1_to_key0=None):
        '''TODO

        '''
        cell_methods0 = self.constructs(construct_type='cell_method')
        cell_methods1 = other.constructs(construct_type='cell_method')

        if len(cell_methods0) != len(cell_methods1):
            if verbose:
                print(
"Verbose: Different numbers of cell methods: {0!r} != {1!r}".format(
    cell_methods0, cell_methods1))
            return False
        
#        axis0_to_axis1 = {}
#        for axis0, axis1 in axis1_to_axis0.items():
#            axis0_to_axis1[axis0] = axis1
            
        axis0_to_axis1 = {axis0: axis1
                          for axis0, axis1 in axis1_to_axis0.items()}
            
        for cm0, cm1 in zip(tuple(cell_methods0.values()),
                            tuple(cell_methods1.values())):
            
            # Check that there are the same number of axes
            axes0 = cm0.get_axes(())
            axes1 = list(cm1.get_axes(()))
            if len(axes0) != len(axes1):
                if verbose:
                    print(
"{0}: Different cell methods (mismatched axes): {1!r}, {2!r}".format(
    cm0.__class__.__name__, cell_methods0, cell_methods1))
                return False
    
            indices = []
            for axis0 in axes0:
                if axis0 is None:
                    return False
                for axis1 in axes1:
                    if axis0 in axis0_to_axis1 and axis1 in axis1_to_axis0:
                        if axis1 == axis0_to_axis1[axis0]:
                            axes1.remove(axis1)
                            indices.append(cm1.get_axes(()).index(axis1))
                            break
                    elif axis0 in axis0_to_axis1 or axis1 in axis1_to_axis0:
                        if verbose:
                            print(
"Verbose: Different cell methods (mismatched axes): {0!r}, {1!r}".format(
    cell_methods0, cell_methods1))
                        return False
                    elif axis0 == axis1:
                        # Assume that the axes are standard names
                        axes1.remove(axis1)
                        indices.append(cm1.get_axes(()).index(axis1))
                    elif axis1 is None:
                        if verbose:
                            print(
"Verbose: Different cell methods (mismatched axes): {0!r}, {1!r}".format(
    cell_methods0, cell_methods1))
                        return False
            #--- End: for

            if len(cm1.get_axes(())) != len(indices):
                if verbose:
                    print("Field: Different cell methods: {0!r}, {1!r}".format(
                        cell_methods0, cell_methods1))
                return False

            cm1 = cm1.sorted(indices=indices)
            cm1.set_axes(axes0)

            if not cm0.equals(cm1, atol=atol, rtol=rtol,
                              verbose=verbose,
                              ignore_type=ignore_type):
                if verbose:
                    print(
"Verbose: Different cell methods: {0!r}, {1!r}".format(
    cell_methods0, cell_methods1))
                return False                
        #--- End: for

        return True
    #--- End: def

    def _equals_domain_axis(self, other, rtol=None, atol=None,
                            verbose=False, ignore_type=False,
                            axis1_to_axis0=None, key1_to_key0=None):
        '''TODO
        '''
        # ------------------------------------------------------------
        # Domain axes
        # ------------------------------------------------------------
        self_sizes  = [d.get_size()
                       for d in self.constructs(construct_type='domain_axis').values()]
        other_sizes = [d.get_size()
                       for d in other.constructs(construct_type='domain_axis').values()]
        
        if sorted(self_sizes) != sorted(other_sizes):
            # There is not a 1-1 correspondence between axis sizes
            if verbose:
                print("{0}: Different domain axes: {1} != {2}".format(
                    self.__class__.__name__,
                    sorted(self.values()),
                    sorted(other.values())))
            return False

        return True
    #--- End: def
    
#--- End: class



from __future__ import print_function
from builtins import (range, super, zip)

import itertools

import numpy
import netCDF4

from .. import core
from .. import mixin

from ..constants  import masked

from . import abstract
from . import NumpyArray


class Data(mixin.Container, core.Data):
    '''An orthogonal multidimensional array with masked values and units.

.. versionadded:: 1.7.0

    '''
    def __init__(self, array=None, units=None, calendar=None,
                 fill_value=None, source=None, copy=True,
                 _use_array=True):
        '''**Initialization**

:Parameters:

    array: numpy array-like or subclass of `Array`, optional
        The array of values. Ignored if the *source* parameter is set.

        *Parameter example:*
          ``array=[34.6]``

        *Parameter example:*
          ``array=[[1, 2], [3, 4]]``

        *Parameter example:*
          ``array=numpy.ma.arange(10).reshape(2, 1, 5)``

    units: `str`, optional
        The physical units of the data. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``units='km hr-1'``

        *Parameter example:*
          ``units='days since 2018-12-01'``

        The units may also be set after initialisation with the
        `set_units` method.

    calendar: `str`, optional
        The calendar for reference time units. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``calendar='360_day'``
        
        The calendar may also be set after initialisation with the
        `set_calendar` method.

    fill_value: optional 
        The fill value of the data. By default, or if None, the numpy
        fill value appropriate to the array's data type will be used.
        TODO. Ignored if the *source* parameter is set.

        *Parameter example:*
          ``fill_value=-999.``
                
        The fill value may also be set after initialisation with the
        `set_fill_value` method.

    source: optional
        Initialize the array, units, calendar and fill value from
        those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(array=array, units=units, calendar=calendar,
                         fill_value=fill_value, source=source,
                         copy=copy, _use_array=_use_array)
                                   
        # The _HDF_chunks attribute is.... Is either None or a
        # dictionary. DO NOT CHANGE IN PLACE.
#        self._HDF_chunks = {}
    #--- End: def
                 
    def __array__(self, *dtype):
        '''The numpy array interface.

.. versionadded:: 1.7.0

:Parameters:

    dtype: optional
        Typecode or data-type to which the array is cast.

:Returns: 

    `numpy.ndarray`
        An independent numpy array of the data.

**Examples:**

>>> import numpy
>>> d = cfdm.Data([1, 2, 3])
>>> a = numpy.array(d)
>>> type(a)
numpy.ndarray
>>> a[0] = -99
>>> d
<Data(3): [1, 2, 3]>
>>> b = numpy.array(d, float)
>>> print(b)
[ 1.  2.  3.]

       '''
        array = self.get_array()
        if not dtype:
            return array
        else:
            return array.astype(dtype[0], copy=False)
    #--- End: def
                  
    def __data__(self):
        '''TODO

Return self

        '''
        return self
    #--- End: def
 
    def __repr__(self):
        '''Called by the `repr` built-in function.

x.__repr__() <==> repr(x)

        '''
        try:        
            shape = self.shape
        except AttributeError:
            shape = ''
        else:
            shape = str(shape)
            shape = shape.replace(',)', ')')
            
        return '<{0}{1}: {2}>'.format(self.__class__.__name__, shape, str(self))
    #--- End: def
   
    def __getitem__(self, indices):
        '''Return a subspace of the data defined by indices

d.__getitem__(indices) <==> d[indices]

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

.. versionadded:: 1.7.0

.. seealso:: `__setitem__`, `_parse_indices`

:Returns:

    `Data`
        The subspace of the data.

**Examples:**

>>> import numpy
>>> d = cfdm.Data(numpy.arange(100, 190).reshape(1, 10, 9))
>>> d.shape
(1, 10, 9)
>>> d[:, :, 1].shape
(1, 10, 1)
>>> d[:, 0].shape
(1, 1, 9)
>>> d[..., 6:3:-1, 3:6].shape
(1, 3, 3)
>>> d[0, [2, 9], [4, 8]].shape
(1, 2, 2)
>>> d[0, :, -2].shape
(1, 10, 1)

        '''
        indices = tuple(self._parse_indices(indices))

        array = self._get_Array(None)
        if array is None:
            raise ValueError("No array!!")
            
        array = array[indices]

        out = self.copy(array=False)
        out._set_Array(array, copy=False)

        return out
    #--- End: def

    def __int__(self):
        '''x.__int__() <==> int(x)

        '''
        if self.size != 1:
            raise TypeError(
"only length-1 arrays can be converted to Python scalars. Got {}".format(self))

        return int(self.get_array())
    #--- End: def

    def __setitem__(self, indices, value):
        '''Assign to data elements defined by indices.

d.__setitem__(indices, x) <==> d[indices]=x

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

**Broadcasting**

The value, or values, being assigned must be broadcastable to the
shape defined by the indices, using the numpy broadcasting rules.

**Missing data**

Data array elements may be set to missing values by assigning them to
`numpy.ma.masked`. Missing values may be unmasked by assigning them to
any other value.

.. versionadded:: 1.7.0

.. seealso:: `__getitem__`, `_parse_indices`

:Returns:

    `None`

**Examples:**

>>> import numpy
>>> d = cfdm.Data(numpy.arange(100, 190).reshape(1, 10, 9))
>>> d.shape
(10, 9)
>>> d[:, :, 1] = -10
>>> d[:, 0] = range(9)
>>> d[..., 6:3:-1, 3:6] = numpy.arange(-18, -9).reshape(3, 3)
>>> d[0, [2, 9], [4, 8]] =  cfdm.Data([[-2, -3]])
>>> d[0, :, -2] = numpy.ma.masked

        '''
        indices = self._parse_indices(indices)
                
        array = self.get_array()

        if value is masked or numpy.ma.isMA(value):
            # The data is not masked but the assignment is masking
            # elements, so turn the non-masked array into a masked
            # one.
            array = array.view(numpy.ma.MaskedArray)        

        self._set_subspace(array, indices, numpy.asanyarray(value))

        self._set_Array(array, copy=False)
    #--- End: def

    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

        '''
        units    = self.get_units(None)
        calendar = self.get_calendar(None)

        if units is not None:
            isreftime = ('since' in units)
        else:
            isreftime = False

        try:
            first = self.first_element()
        except:            
            out = ''
            if units:
                out += ' {0}'.format(units)
            if calendar:
                out += ' {0}'.format(calendar)
               
            return out
        #--- End: try

        size = self.size
        ndim = self.ndim
        open_brackets  = '[' * ndim
        close_brackets = ']' * ndim

        if size == 1:
            if isreftime:
                # Convert reference time to date-time
                try:
                    first = type(self)(
                        numpy.ma.array(first), units, calendar).get_dtarray()
                except (ValueError, OverflowError):
                    first = '??'

            out = '{0}{1}{2}'.format(open_brackets,
                                     first,
                                     close_brackets)
        else:
            last = self.last_element()
            if isreftime:
                # Convert reference times to date-times
                try:
                    first, last = type(self)(
                        numpy.ma.array([first, last]), units, calendar).get_dtarray()
                except (ValueError, OverflowError):
                    first, last = ('??', '??')

            if size > 3:
                out = '{0}{1!s}, ..., {2!s}{3}'.format(open_brackets,
                                                       first, last,
                                                       close_brackets)
            elif size == 3:                
                middle = self.second_element()
                if isreftime:
                    # Convert reference time to date-time
                    try:
                        middle = type(self)(
                            numpy.ma.array(middle), units, calendar).get_dtarray()
                    except (ValueError, OverflowError):
                        middle = '??'
                        
                out = '{0}{1!s}, {2!s}, {3!s}{4}'.format(open_brackets,
                                                         first, middle, last,
                                                         close_brackets)
            else:
                out = '{0}{1!s}, {2!s}{3}'.format(open_brackets,
                                                  first, last,
                                                  close_brackets)
        #--- End: if
        
        if isreftime:
            if calendar:
                out += ' {0}'.format(calendar)
        elif units:
            out += ' {0}'.format(units)
            
        return out
    #--- End: def

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _item(self, index):
        '''Copy an element of an array to a standard Python scalar and return it.

It is assumed, but not checked, that the given index selects exactly
one element.

**Examples:**

>>> import numpy
>>> d = Data([[1, 2, 3]], 'km')
>>> x = d._item((0, -1))
>>> print(x, type(x))
3 <type 'int'>
>>> x = d._item(1)
>>> print(x, type(x))
2 <type 'int'>
>>> d[0, 1] = numpy.ma.masked
>>> d._item((slice(None), slice(1, 2)))
masked

        '''
        array = self[index].get_array()

        if not numpy.ma.isMA(array):
            return array.item()

        mask = array.mask
        if mask is numpy.ma.nomask or not mask.item():
            return array.item()

        return numpy.ma.masked
    #--- End: def
    
    def _parse_axes(self, axes):
        '''TODO

:Parameters:

    axes: (sequence of) `int`
        The axes of the data. May be one of, or a sequence of any
        combination of zero or more of:

          * The integer position of a dimension in the data (negative
            indices allowed).

:Returns:

    `tuple`

**Examples:**

        '''
        ndim = self.ndim

        if isinstance(axes, int):
            axes = (axes,)
            
        axes2 = []
        for axis in axes:
            if 0 <= axis < ndim:
                axes2.append(axis)
            elif -ndim <= axis < 0:
                axes2.append(axis + ndim)
            else:
                raise ValueError(
                    "Invalid axis: {!r}".format(axis))
        #--- End: for
            
        # Check for duplicate axes
        n = len(axes2)
        if n > len(set(axes2)) >= 1:
            raise ValueError("Duplicate axis: {}".format(axes2))
        
        return tuple(axes2)
    #--- End: def

    def _set_Array(self, array, copy=True):
        '''Set the array.

:Parameters:

    array: numpy array-like or subclass of `Array`, optional
        The array to be inserted.

:Returns:

    `None`

**Examples:**

>>> d._set_Array(a)

        '''
        if not isinstance(array, abstract.Array):
            if not isinstance(array, numpy.ndarray):
                array = numpy.asanyarray(array)
                
            array = NumpyArray(array)

        super()._set_Array(array, copy=copy)
    #--- End: def

    @classmethod
    def _set_subspace(cls, array, indices, value):
        '''TODO
        '''
        axes_with_list_indices = [i for i, x in enumerate(indices)
                                  if not isinstance(x, slice)]

        if len(axes_with_list_indices) < 2: 
            # ------------------------------------------------------------
            # At most one axis has a list-of-integers index so we can do a
            # normal numpy assignment
            # ------------------------------------------------------------
            array[tuple(indices)] = value
        else:
            # ------------------------------------------------------------
            # At least two axes have list-of-integers indices so we can't
            # do a normal numpy assignment
            # ------------------------------------------------------------
            indices1 = indices[:]
            for i, x in enumerate(indices):
                if i in axes_with_list_indices:
                    # This index is a list of integers
                    y = []
                    args = [iter(x)] * 2
                    for start, stop in itertools.zip_longest(*args):
                        if not stop:
                            y.append(slice(start, start+1))
                        else:
                            step = stop - start
                            stop += 1
                            y.append(slice(start, stop, step))
                    #--- End: for
                    indices1[i] = y
                else:
                    indices1[i] = (x,)
            #--- End: for

            if numpy.size(value) == 1:
                for i in itertools.product(*indices1):
                    array[i] = value
                    
            else:
                indices2 = []
                ndim_difference = array.ndim - numpy.ndim(value)
                for i, n in enumerate(numpy.shape(value)):
                    if n == 1:
                        indices2.append((slice(None),))
                    elif i + ndim_difference in axes_with_list_indices:
                        y = []
                        start = 0
                        while start < n:
                            stop = start + 2
                            y.append(slice(start, stop))
                            start = stop
                        #--- End: while
                        indices2.append(y)
                    else:
                        indices2.append((slice(None),))
                #--- End: for

                for i, j in zip(itertools.product(*indices1), itertools.product(*indices2)):
                    array[i] = value[j]
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, array=True):
        '''Return a deep copy.

``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

:Parameters:

    array: `bool`, optional
        If False then do not copy the array. By default the array is
        copied.

:Returns:

        The deep copy.

**Examples:**

>>> e = d.copy()
>>> e = d.copy(array=False)

        '''
        new = super().copy(array=array)
#        new.HDF_chunks(self.HDF_chunks())
        return new
    #--- End: def

    def insert_dimension(self, position=0):
        '''Expand the shape of the data array.

Inserts a new size 1 axis, corresponding to a given position in the
data array shape.

.. versionadded:: 1.7.0

.. seealso:: `squeeze`, `transpose`

:Parameters:

    position: `int`, optional
        Specify the position that the new axis will have in the data
        array. By default the new axis has position 0, the slowest
        varying position. Negative integers counting from the last
        position are allowed.

        *Parameter example:*
          ``position=2``

        *Parameter example:*
          ``position=-1``

:Returns:

    `Data`
        The new data array with expanded data axes.

**Examples:**

>>> d.shape
(19, 73, 96)
>>> d.insert_dimension('domainaxis3').shape
(1, 96, 73, 19)
>>> d.insert_dimension('domainaxis3', position=3).shape
(19, 73, 96, 1)
>>> d.insert_dimension('domainaxis3', position=-1).shape
(19, 73, 1, 96)

        '''
        # Parse position
        ndim = self.ndim 
        if -ndim-1 <= position < 0:
            position += ndim + 1
        elif not 0 <= position <= ndim:
            raise ValueError(
                "Can't expand dimension sof data: Invalid position: {!r}".format(position))

        array = self.get_array()
        array = numpy.expand_dims(array, position)

        d = self.copy(array=False)
        d._set_Array(array, copy=False)

#        if d._HDF_chunks:            
#            HDF = {}
#            for axis in axes:
#                HDF[axis] = None
#
#            d.HDF_chunks(HDF)

        return d
    #--- End: def

#    def compress_by_gathering(self, list_data, compressed_axes, replace_list_data=False

#    def set_list_data(self, list_data, compressed_axes=None,
#                      copy=True):
#        '''
#        '''
#        compression_type = self.compression_type
#        if compression_type == 'gathered':
#            self._set_Array().set_list_data(list_data)
##            raise ValueError("eqweqweweqw 1")
#        elif compression_type:
#            raise ValueError("eqweqweweqw 2")
#
#        compressed_array = self._compress_by_gathering(list_data, compressed_axes)
#
#        self._set_Array(GatheredArray(compressed_array=compressed_array,
#                                      shape=self.shape,
#                                      size=self.size, ndim=self.ndim,
#                                      sample_axis=compressed_axes[0],
#                                      list_array=list_data))
#    #--- End: def

    def get_dtarray(self):
        '''An independent numpy array of date-time objects.

Only applicable for reference time units.

If the calendar has not been set then the CF default calendar of
"gregorian" will be used.

Conversions are carried out with the `netCDF4.num2date` function.

.. versionadded:: 1.7.0

.. seealso:: `get_array`

**Examples:**

>>> d.get_units()
'days since 2018-12-01'
>>> d.get_calendar()
'360_day'
>>> d.get_array()
30.0
>>> print(d.get_dtarray())
[2019-01-01 00:00:00]

        '''
        array = self.get_array()

        mask = None
        if numpy.ma.isMA(array):
            # num2date has issues if the mask is nomask
            mask = array.mask
            if mask is numpy.ma.nomask or not numpy.ma.is_masked(array):
                mask = None
                array = array.view(numpy.ndarray)
        #--- End: if
        
        array = netCDF4.num2date(array, units=self.get_units(None),
                                 calendar=self.get_calendar('standard'),
                                 only_use_cftime_datetimes=True)

        if mask is None:
            # There is no missing data
            array = numpy.array(array, dtype=object)
        else:
            # There is missing data
            array = numpy.ma.masked_where(mask, array)
            if not numpy.ndim(array):
                array = numpy.ma.masked_all((), dtype=object)

        return array
    #--- End: def

    def get_count_variable(self, *default):
        '''Return the count variable for a compressed array.

.. versionadded:: 1.7.0

.. seealso:: `get_index_variable`, `get_list_variable`

:Parameters:

    default: optional
        Return *default* if a count variable has not been set.

:Returns:

        The count variable. If unset then *default* is returned, if
        provided.

**Examples:**

>>> c = d.get_count_variable(None)

        '''
        array = self._get_Array(None)
        if array is None:
            if default:
                return default

            raise AttributeError("{!r} has no count variable".format(
                self.__class__.__name__))

        variable = array.get_count_variable(None)
        if variable is None:
            if default:
                return default[0]

            raise AttributeError("{!r} has no count variable".format(
                self.__class__.__name__))
        
        return variable
    #--- End: def

    def get_index_variable(self, *default):
        '''Return the index variable for a compressed array.

.. versionadded:: 1.7.0

.. seealso:: `get_count_variable`, `get_list_variable`

:Parameters:

    default: optional
        Return *default* if index variable has not been set.

:Returns:

        The index variable. If unset then *default* is returned, if
        provided.

**Examples:**

>>> i = d.get_index_variable(None)

        '''
        array = self._get_Array(None)
        if array is None:
            if default:
                return default

            raise AttributeError("{!r} has no index variable".format(
                self.__class__.__name__))

        variable = array.get_index_variable(None)
        if variable is None:
            if default:
                return default[0]

            raise AttributeError("{!r} has no index variable".format(
                self.__class__.__name__))
        
        return variable
    #--- End: def

    def get_list_variable(self, *default):
        '''Return the list variable for a compressed array.

.. versionadded:: 1.7.0

.. seealso:: `get_count_variable`, `get_index_variable`

:Parameters:

    default: optional
        Return *default* if a list variable has not been set.

:Returns:

        The list variable. If unset then *default* is returned, if
        provided.

**Examples:**

>>> l = d.get_list_variable(None)

        '''
        array = self._get_Array(None)
        if array is None:
            if default:
                return default

            raise AttributeError("{!r} has no list variable".format(
                self.__class__.__name__))

        variable = array.get_list_variable(None)
        if variable is None:
            if default:
                return default[0]

            raise AttributeError("{!r} has no list variable".format(
                self.__class__.__name__))
        
        return variable
    #--- End: def

    def get_compressed_dimension(self, *default):
        '''Return the position of the compressed dimension in the compressed
array.

. versionadded:: 1.7.0

.. seealso:: `get_compressed_axearray`, `get_compressed_axes`,
             `get_compressed_type`

:Parameters:

    default: optional
        Return *default* if the underlying array is not compressed.

:Returns:

    `int`
        The position of the compressed dimension in the compressed
        array. If the underlying is not compressed then *default* is
        returned, if provided.

**Examples:**

>>> i = d.get_compressed_dimension()

        '''        
        a = self._get_Array(None)

        compressed_dimension = a.get_compressed_dimension()
        if compressed_dimension is None:
            raise ValueError("not compressed: can't get compressed dimension")

        return compressed_dimension
    #--- End: def

    
    def _parse_indices(self, indices):
        '''TODO
    
:Parameters:
    
    indices: `tuple` (not a `list`!)
    
:Returns:
    
    `list`
    
**Examples:**
    
    '''
        shape = self.shape
        
        parsed_indices = []
    
        if not isinstance(indices, tuple):
            indices = (indices,)
    
        # Initialize the list of parsed indices as the input indices with any
        # Ellipsis objects expanded
        length = len(indices)
        n = len(shape)
        ndim = n
        for index in indices:
            if index is Ellipsis:
                m = n - length + 1
                parsed_indices.extend([slice(None)] * m)
                n -= m            
            else:
                parsed_indices.append(index)
                n -= 1
    
            length -= 1
        #--- End: for
        len_parsed_indices = len(parsed_indices)
    
        if ndim and len_parsed_indices > ndim:
            raise IndexError(
                "Invalid indices for data with shape {}: {} ".format(
                    shape, parsed_indices))
    
        if len_parsed_indices < ndim:
            parsed_indices.extend([slice(None)]*(ndim-len_parsed_indices))
    
        if not ndim and parsed_indices:
            raise IndexError(
                "Scalar data can only be indexed with () or Ellipsis")
    
        for i, (index, size) in enumerate(zip(parsed_indices, shape)):
            if isinstance(index, slice):            
                continue
    
            if isinstance(index, int):
                # E.g. 43 -> slice(43, 44, 1)
                if index < 0: 
                    index += size
    
                index = slice(index, index+1, 1)
            else:
                if getattr(getattr(index, 'dtype', None), 'kind', None) == 'b':
                    # E.g. index is [True, False, True] -> [0, 2]
                    #
                    # Convert booleans to non-negative integers. We're
                    # assuming that anything with a dtype attribute also
                    # has a size attribute.
                    if index.size != size:
                        raise IndexError(
"Invalid indices for data with shape {}: {} ".format(
                                shape, parsed_indices))

                    index = numpy.where(index)[0]
                #--- End: if
    
                if not numpy.ndim(index):
                    if index < 0:
                        index += size
    
                    index = slice(index, index+1, 1)
                else:
                    len_index = len(index)
                    if len_index == 1:
                        # E.g. [3] -> slice(3, 4, 1)
                        index = index[0]
                        if index < 0:
                            index += size
                        
                        index = slice(index, index+1, 1)
                    else:
                        # E.g. [1, 3, 4] -> [1, 3, 4]
                        pass
            #--- End: if
            
            parsed_indices[i] = index    
        #--- End: for
    
        return parsed_indices
    #--- End: def

    def max(self, axes=None):
        '''Return the maximum of an array or the maximum along axes.

Missing data array elements are omitted from the calculation.

.. seealso:: `min`

:Parameters:

    axes: (sequence of) `int`, optional

:Returns:

    `Data`
        Maximum of the data along the specified axes.

**Examples:**

        '''
        # Parse the axes. By default flattened input is used.
        if axes is not None:
            try:
                axes = self._parse_axes(axes)
            except ValueError as error:
                raise ValueError("Can't find maximum of data: {}".format(error))
        #--- End: if
        
        array = self.get_array()
        array = numpy.amax(array, axis=axes, keepdims=True)

        d = self.copy(array=False)
        d._set_Array(array, copy=False)
        
#        if d._HDF_chunks:            
#            HDF = {}
#            for axis in axes:
#                HDF[axis] = None
#
#            d.HDF_chunks(HDF)

        return d
    #--- End: def

    def min(self, axes=None):
        '''Return the minimum of an array or minimum along axes.

Missing data array elements are omitted from the calculation.

.. seealso:: `max`

:Parameters:

    axes: (sequence of) `int`, optional

:Returns:

    `Data`
        Minimum of the data along the specified axes.

**Examples:**

        '''            
        # Parse the axes. By default flattened input is used.
        if axes is not None:
            try:
                axes = self._parse_axes(axes)
            except ValueError as error:
                raise ValueError("Can't find minimum of data: {}".format(error))
        #--- End: if

        array = self.get_array()
        array = numpy.amin(array, axis=axes, keepdims=True)

        d = self.copy(array=False)
        d._set_Array(array, copy=False)

#        if d._HDF_chunks:            
#            HDF = {}
#            for axis in axes:
#                HDF[axis] = None
#
#            d.HDF_chunks(HDF)

        return d
    #--- End: def

#    def get_HDF_chunks(self, dddd):
#        '''Set HDF5 chunks for the data array.
#    
#Chunking refers to a storage layout where the data array is
#partitioned into fixed-size multi-dimensional chunks when written to a
#netCDF4 file on disk. Chunking is ignored if the data array is written
#to a netCDF3 format file.
#    
#A chunk has the same rank as the data array, but with fewer (or no
#more) elements along each axis. The chunk is defined by a dictionary
#in which each key identifies an axis (by its index in the data array
#shape) and its value is the chunk size (i.e. number of axis elements)
#for that axis.
#    
#If a given chunk size for an axis is larger than the axis size, then
#the size of the axis at the time of writing to disk will be used
#instead.
#    
#If chunk sizes have been specified for some but not all axes, then the
#each unspecified chunk size is assumed to be the full size of its
#axis.
#
#If no chunk sizes have been set for any axes then the netCDF default
#chunk is used. See
#http://www.unidata.ucar.edu/software/netcdf/docs/netcdf_perf_chunking.html.
#
#A detailed discussion of HDF chunking and I/O performance is available
#at https://www.hdfgroup.org/HDF5/doc/H5.user/Chunking.html and
#http://www.unidata.ucar.edu/software/netcdf/workshops/2011/nc4chunking. Basically,
#you want the chunks for each dimension to match as closely as possible
#the size and shape of the data block that users will read from the
#file.
#
#        '''
#        self._HDF_chunks = tuplchunks.copy()
#
#        
#    def HDF_chunks(self, *chunks):
#        '''
#        '''
#        _HDF_chunks = self._HDF_chunks
#
#        org_HDF_chunks = dict([(i, _HDF_chunks.get(i))
#                               for i in range(self.ndim)])
#
#        org_HDF_chunks = _HDF_chunks.copy()
#        
#        if not chunks:
#            return org_HDF_chunks
#
##        if _HDF_chunks is None:
##            _HDF_chunks = {}
##        else:
##        _HDF_chunks = _HDF_chunks.copy()
#
##        org_HDF_chunks = _HDF_chunks.copy()
#            
# 
#        if not chunks:
#            return org_HDF_chunks
#
#        chunks = chunks[0]
#        
#        if chunks is None:
#            # Clear all chunking
#            self._HDF_chunks = {}
#            return org_HDF_chunks
#
##        for i in range(self.ndim):
#            
#
#        for axis, size in chunks.iteritems():
#            if size is not None:
#                _HDF_chunks[axis] = size
#            else:
#                _HDF_chunks.pop(axis, None)
#                
##        if _HDF_chunks.values() == [None] * len(_HDF_chunks):
##            _HDF_chunks = None
#
##        self._HDF_chunks = _HDF_chunks
#            
#        return org_HDF_chunks
#    #--- End: def

    def squeeze(self, axes=None):
        '''Remove size 1 axes from the data.

By default all size 1 axes are removed, but particular axes may be
selected with the keyword arguments.

.. versionadded:: 1.7.0

.. seealso:: `insert_dimension`, `transpose`

:Parameters:

    axes: (sequence of) `int`, optional
        The positions of the size one axes to be removed. By default
        all size one axes are removed. Each axis is identified by its
        original integer position. Negative integers counting from the
        last position are allowed.

        *Parameter example:*
          ``axes=0``

        *Parameter example:*
          ``axes=-2``

        *Parameter example:*
          ``axes=[2, 0]``

:Returns:

    `Data`
        The new data array with removed data axes.

**Examples:**

>>> d.shape
(1, 73, 1, 96)
>>> f.squeeze().shape
(73, 96)
>>> d.squeeze(0).shape
(73, 1, 96)
>>> d.squeeze([-3, 2]).shape
(73, 96)

        '''
        d = self.copy()

        if not d.ndim:
            if axes:
                raise ValueError(
"Can't squeeze data: axes {} is not allowed data with shape {}".format(axes, d.shape))

            return d

        shape = d.shape

        if axes is None:
            axes = [i for i, n in enumerate(shape) if n == 1]
        else:
            try:
                axes = self._parse_axes(axes)
            except ValueError as error:
                raise ValueError("Can't squeeze data: {}".format(error))

            # Check the squeeze axes
            for i in axes:
                if shape[i] > 1:
                    raise ValueError(
"Can't squeeze data: Can't remove axis of size {}".format(shape[i]))
        #--- End: if

        if not axes:
            return d

        array = self.get_array()
        array = numpy.squeeze(array, axes)

        d._set_Array(array, copy=False)

        return d
    #--- End: def

    def sum(self, axes=None):
        '''Return the sum of an array or the sum along axes.

Missing data array elements are omitted from the calculation.

.. seealso:: `max`, `min`

:Parameters:

    axes: (sequence of) `int`, optional

:Returns:

    `Data`
        The sum of the data along the specified axes.

**Examples:**

        '''
        # Parse the axes. By default flattened input is used.
        if axes is not None:
            try:
                axes = self._parse_axes(axes)
            except ValueError as error:
                raise ValueError("Can't sum data: {}".format(error))
        #--- End: if
        
        array = self.get_array()
        array = numpy.sum(array, axis=axes, keepdims=True)
            
        d = self.copy()
        d._set_Array(array, copy=False)

#        if d._HDF_chunks:            
#            HDF = {}
#            for axis in axes:
#                HDF[axis] = None
#
#            d.HDF_chunks(HDF)
        
        return d
    #--- End: def

    def transpose(self, axes=None):
        '''Permute the axes of the data array.

.. versionadded:: 1.7.0

.. seealso:: `insert_dimension`, `squeeze`

:Parameters:

    axes: (sequence of) `int`
        The new axis order. By default the order is reversed. Each
        axis in the new order is identified by its original integer
        position. Negative integers counting from the last position
        are allowed.

        *Parameter example:*
          ``axes=[2, 0, 1]``

        **Parameter example:**
          ``axes=[-1, 0, 1]``

:Returns:

    `Data`
        The new data array with permuted data axes.

**Examples:**

>>> d.shape
(19, 73, 96)
>>> d.tranpose().shape
(96, 73, 19)
>>> d.tranpose([1, 0, 2]).shape
(73, 19, 96)

        '''
        d = self.copy()
        
        ndim = d.ndim    
        
        # Parse the axes. By default, reverse the order of the axes.
        if axes is None:
            if ndim <= 1:
                return d

            axes = list(range(ndim-1, -1, -1))
        else:
            try:
                axes = self._parse_axes(axes)
            except ValueError as error:
                raise ValueError("Can't transpose data: {}".format(error))

            # Return unchanged if axes are in the same order as the data
            if axes == list(range(ndim)):
                return d

            if len(axes) != ndim:
                raise ValueError(
                    "Can't transpose data: Axes don't match array: {}".format(axes))
        #--- End: if

        array = self.get_array()
        array = numpy.transpose(array, axes=axes)

        d._set_Array(array, copy=False)

        return d
    #--- End: def

    def get_compressed_array(self):
        '''Return an independent numpy array containing the compressed data.

.. versionadded:: 1.7.0

.. seealso:: `get_compressed_axes`, `get_compressed_dimension`,
             `get_compression_type`,

:Returns:

     `numpy.ndarray`
        An independent numpy array of the compressed data.

**Examples:**

>>> a = d.get_compressed_array()

        '''
        ca = self._get_Array(None)

        if not ca.get_compressed_axes():
            raise ValueError("not compressed: can't get compressed array")

        return ca.get_compressed_array()
    #--- End: def

    def get_compressed_axes(self):
        '''Return the dimensions that are compressed.

.. versionadded:: 1.7.0

:Returns:

    `list`
        TODO The axes of the data that are compressed to a single axis
        in the internal array.

**Examples:**

TODO

        '''
        ca = self._get_Array(None)

        if ca is None:
            return []

        return ca.get_compressed_axes()
    #--- End: def

    def get_compression_type(self):
        '''Return the type of compression applied to the underlying array.

.. versionadded:: 1.7.0

.. seealso:: `get_compressed_array`, `compression_axes`,
             `get_compressed_dimension`

:Returns:

    `str` or `None`
        The compression type. An empty string means that no
        compression has been applied.
        
**Examples:**

>>> d.get_compression_type()
''

>>> d.get_compression_type()
'gathered'

>>> d.get_compression_type()
'ragged contiguous'

        '''
        ma = self._get_Array(None)
        if ma is None:
            return ''

        return ma.get_compression_type()
    #--- End: def
    
    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_compression=False, ignore_type=False,
               _check_values=True):
        '''Whether two data arrays are the same.

Equality is strict by default. This means that for data arrays to be
considered equal:

* the units and calendar must be the same,

..

* the fill value must be the same (see the *ignore_fill_value*
  parameter), and

..

* the arrays must have same shape and data type, the same missing data
  mask, and be element-wise equal (see the *ignore_data_type*
  parameter).

Two numerical elements ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

The compression type and, if appliciable, the underlying compressed
arrays must be the same, as well as the arrays in their uncompressed
forms. See the *ignore_compression* parameter.

Any type of object may be tested but, in general, equality is only
possible with another cell measure construct, or a subclass of
one. See the *ignore_type* parameter.

.. versionadded:: 1.7.0

:Parameters:

    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    ignore_fill_value: `bool`, optional
        If True then the fill value is omitted from the comparison.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical data array
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_compression: `bool`, optional
        If True then any compression applied to the underlying arrays
        is ignored and only the uncompressed arrays are tested for
        equality. By default the compression type and, if appliciable,
        the underlying compressed arrays must be the same, as well as
        the arrays in their uncompressed forms

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another data array, or a subclass of
        one. If *ignore_type* is True then then ``Data(source=other)``
        is tested, rather than the ``other`` defined by the *other*
        parameter.

:Returns: 
  
    `bool`
        Whether the two data arrays are equal.

**Examples:**

>>> d.equals(d)
True
>>> d.equals(d.copy())
True
>>> d.equals('not a data array')
False

        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp in (True, False):
            return pp
        
        other = pp
        
        # Check that each instance has the same shape
        if self.shape != other.shape:
            if verbose:
                print("{0}: Different shapes: {1} != {2}".format(
                    self.__class__.__name__, self.shape, other.shape))
            return False

        # Check that each instance has the same units
        for attr in ('units', 'calendar'):
            x = getattr(self, 'get_'+attr)(None)
            y = getattr(other, 'get_'+attr)(None)
            if x != y:
                if verbose:
                    print("{0}: Different {1}: {2!r} != {3!r}".format(
                        self.__class__.__name__, attr, x, y))
                return False
        #--- End: for
           
        # Check that each instance has the same fill value
        if (not ignore_fill_value and
            self.get_fill_value(None) != other.get_fill_value(None)):
            if verbose:
                print("{0}: Different fill value: {1} != {2}".format(
                    self.__class__.__name__, 
                    self.get_fill_value(None), other.get_fill_value(None)))
            return False

        # Check that each instance has the same data type
        if not ignore_data_type and self.dtype != other.dtype:
            if verbose:
                print("{0}: Different data types: {1} != {2}".format(
                    self.__class__.__name__, self.dtype, other.dtype))
            return False

        # Return now if we have been asked to not check the array
        # values
        if not _check_values:
            return True

        if not ignore_compression:
            # --------------------------------------------------------
            # Check for equal compression types
            # --------------------------------------------------------
            compression_type = self.get_compression_type()
            if compression_type != other.get_compression_type():
                if verbose:
                    print("{0}: Different compression types: {1} != {2}".format(
                        self.__class__.__name__,
                        compression_type,
                        other.get_compression_type()))
                return False
            
            # --------------------------------------------------------
            # Check for equal compressed array values
            # --------------------------------------------------------
            if compression_type:
                if not self._equals(self.get_compressed_array(),
                                    other.get_compressed_array(),
                                    rtol=rtol, atol=atol):
                    if verbose:
                        print("{0}: Different compressed array values".format(
                            self.__class__.__name__))
                    return False
        #--- End: if
        
        # ------------------------------------------------------------
        # Check for equal (uncompressed) array values
        # ------------------------------------------------------------
        if not self._equals(self.get_array(), other.get_array(),
                            rtol=rtol, atol=atol):
            if verbose:
                print("{0}: Different array values".format(
                    self.__class__.__name__))
            return False

        # ------------------------------------------------------------
        # Still here? Then the two data arrays are equal.
        # ------------------------------------------------------------
        return True            
    #--- End: def

    def first_element(self):
        '''TODO
        '''
        return self._item((slice(0, 1),)*self.ndim)
    #--- End: def
    
    def last_element(self):
        '''TODO
        '''
        
        return self._item((slice(-1, None),)*self.ndim)
    #--- End: def
 
    def second_element(self):
        '''TODO
        '''
        return self._item((slice(0, 1),)*(self.ndim-1) + (slice(1, 2),))
    #--- End: def

#    def set_dtype(self, value):
#        '''
#        '''
#        value = numpy.dtype(value)
 #       if value != self.dtype:
#            array = numpy.asanyarray(self.get_array(), dtype=value)
#            self._set_Array(array, copy=False)
#    #--- End: def

    def unique(self):
        '''The unique elements of the data.

The unique elements are sorted into a one dimensional array. with no
missing values.

.. versionadded:: 1.7.0

:Returns:

    `Data`
        The unique elements.

**Examples:**

>>> d = Data([[4, 2, 1], [1, 2, 3]], 'metre')
>>> d.unique()
<Data(4): [1, 2, 3, 4] metre>
>>> d[1, -1] = masked
>>> d.unique()
<Data(3): [1, 2, 4] metre>

        '''
        array = self.get_array()
        array = numpy.unique(array)

        if numpy.ma.is_masked(array):
            array = array.compressed()

        out = self.copy(array=False)
        out._set_Array(array, copy=False)

        return out
    #--- End: def

#--- End: class



from builtins import object
from future.utils import with_metaclass
import abc


class IO(with_metaclass(abc.ABCMeta, object)):
    '''Abstract base class for reading and writing Fields.

    '''
    def __init__(self, implementation):
        '''**Initialisation**

:Parameters:

    implementation: `Implementation'
        The objects required to represent a Field.

        '''
        self.implementation = implementation
    #--- End: def

    @abc.abstractmethod
    def file_close(self, *args, **kwargs):
        '''Close the dataset file.
        '''
        raise NotImplementedError()
    #--- End: def

    @abc.abstractmethod
    def file_open(self, *args, **kwargs):
        '''Open the dataset file.
        '''
        raise NotImplementedError()
    #--- End: def

#--- End: class


class IORead(with_metaclass(abc.ABCMeta, IO)):
    '''Abstract base class for instantiating Fields from a dataset.

    '''
    @abc.abstractmethod
    def read(self, *args, **kwargs):
        '''Read fields from a netCDF file.

        '''
        raise NotImplementedError()
    #--- End: def

#--- End: class


class IOWrite(with_metaclass(abc.ABCMeta, IO)):
    '''Abstract base class for writing Fields to a dataset.

    '''
    @abc.abstractmethod
    def write(self, *args, **kwargs):
        '''Write fields to a netCDF file.

        '''
        raise NotImplementedError()
    #--- End: def
    
#--- End: class



from __future__ import print_function
import datetime
import inspect
import os
import tempfile
import unittest

import numpy

import cfdm

class StringTest(unittest.TestCase):
    def setUp(self):
        self.test_only = []

        (fd, self.tempfilename) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_', dir='.')
        os.close(fd)        
    #--- End: def

    def tearDown(self):
        os.remove(self.tempfilename)
    #--- End: def
        
    def test_STRING(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for array in (numpy.ma.array(list('abcdefghij'), dtype='S'),
                      numpy.ma.array(['a', 'b1', 'c12', 'd123', 'e1234', 'f', 'g', 'h', 'i', 'j'], dtype='S')):

            # Initialize the field
            tas = cfdm.Field(
                properties={'project': 'research',
                            'standard_name': 'air_temperature',
                            'units': 'K'})
            
            # Create and set domain axes
            axis_T = tas.set_construct(cfdm.DomainAxis(1))
            axis_Z = tas.set_construct(cfdm.DomainAxis(1))
            axis_Y = tas.set_construct(cfdm.DomainAxis(10))
            axis_X = tas.set_construct(cfdm.DomainAxis(9))
            
            # Set the field data
            tas.set_data(cfdm.Data(numpy.arange(90.).reshape(10, 9)),
                         axes=[axis_Y, axis_X])
            
            # Create and set the dimension coordinates
            dimension_coordinate_Y = cfdm.DimensionCoordinate(
                properties={'standard_name': 'grid_latitude',
                            'units': 'degrees'},
                data=cfdm.Data(numpy.arange(10.)),
                bounds=cfdm.Bounds(data=cfdm.Data(numpy.arange(20).reshape(10, 2))))
            
            dimension_coordinate_X = cfdm.DimensionCoordinate(
                properties={'standard_name': 'grid_longitude',
                        'units': 'degrees'},
                data=cfdm.Data(numpy.arange(9.)),
                bounds=cfdm.Bounds(data=cfdm.Data(numpy.arange(18).reshape(9, 2))))
            
            dim_Y = tas.set_construct(dimension_coordinate_Y, axes=[axis_Y])
            dim_X = tas.set_construct(dimension_coordinate_X, axes=[axis_X])
            
            # Create and set the auxiliary coordinates
#            array = numpy.ma.array(list('abcdefghij'), dtype='S')
            array[0] = numpy.ma.masked
            aux0 = cfdm.AuxiliaryCoordinate(properties={'long_name': 'Grid latitude name'},
                                            data=cfdm.Data(array))
            
            tas.set_construct(aux0, axes=[axis_Y])
    
    
            cfdm.write(tas, self.tempfilename)
    
            tas1 = cfdm.read(self.tempfilename)[0]
    
            aux1 = tas1.get_construct('long_name:Grid latitude name')
            self.assertTrue(aux0.data.shape == array.shape, aux0.data.shape)
            self.assertTrue(aux1.data.shape == array.shape, aux1.data.shape)
        #--- End: for
    #--- End: def
    
#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    unittest.main(verbosity=2)



from builtins import (range, super, zip)

import numpy

from . import abstract
from . import mixin


class RaggedIndexedContiguousArray(mixin.RaggedContiguous,
                                   mixin.RaggedIndexed,
                                   abstract.CompressedArray):
    '''An underlying indexed contiguous ragged array.

A collection of features, each of which is sequence of (vertical)
profiles, stored using an indexed contiguous ragged array combines all
feature elements along a single dimension (the "sample dimension")
such that a contiguous ragged array representation is used for each
profile and the indexed ragged array representation to organise the
profiles into timeseries.

The information needed to uncompress the data is stored in a "count
variable" that gives the size of each profile; and in a "index
variable" that specifies the feature that each profile belongs to.

.. versionadded:: 1.7.0

    '''
    def __init__(self, compressed_array=None, shape=None, size=None,
                 ndim=None, count_variable=None, index_variable=None):
        '''**Initialization**

:Parameters:

    compressed_array: subclass of `Array`
        The compressed array.

    shape: `tuple`
        The uncompressed array dimension sizes.

    size: `int`
        Number of elements in the uncompressed array.

    ndim: `int`
        The number of uncompressed array dimensions

    count_variable: `Count`
        The count variable required to uncompress the data,
        corresponding to a CF-netCDF count variable.

    index_variable: `Index`
        The index variable required to uncompress the data,
        corresponding to a CF-netCDF CF-netCDF index variable.

        '''
        super().__init__(compressed_array=compressed_array,
                         shape=shape, size=size, ndim=ndim,
                         count_variable=count_variable,
                         index_variable=index_variable,
                         compression_type='ragged indexed contiguous',
                         compressed_dimension=0)    
    #--- End: def

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

Returns an subspace of the uncompressed data an independent numpy
array.

The indices that define the subspace are relative to the uncompressed
data and must be either `Ellipsis` or a sequence that contains an
index for each dimension. In the latter case, each dimension's index
must either be a `slice` object or a sequence of two or more integers.

Indexing is similar to numpy indexing. The only difference to numpy
indexing (given the restrictions on the type of indices allowed) is:

  * When two or more dimension's indices are sequences of integers
    then these indices work independently along each dimension
    (similar to the way vector subscripts work in Fortran).

        '''
        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        
        compressed_array = self._get_compressed_Array()

        # Initialise the un-sliced uncompressed array
        uarray = numpy.ma.masked_all(self.shape, dtype=self.dtype)

        count_array = self.get_count_variable().get_array()
        index_array = self.get_index_variable().get_array()
        
        # Loop over instances
        for i in range(uarray.shape[0]):
            
            # For all of the profiles in ths instance, find the
            # locations in the count array of the number of elements
            # in the profile
            xprofile_indices = numpy.where(index_array == i)[0]
                
            # Find the number of profiles in this instance
            n_profiles = xprofile_indices.size
            
            # Loop over profiles in this instance
            for j in range(uarray.shape[1]):
                if j >= n_profiles:
                    continue
                
                # Find the location in the count array of the number
                # of elements in this profile
                profile_index = xprofile_indices[j]
                
                if profile_index == 0:
                    start = 0
                else:                    
                    start = int(count_array[:profile_index].sum())
                    
                stop = start + int(count_array[profile_index])
                
                sample_indices = slice(start, stop)
                
                u_indices = (i, #slice(i, i+1),
                             j, #slice(j, j+1), 
                             slice(0, stop-start)) #slice(0, sample_indices.stop - sample_indices.start))
                
                uarray[u_indices] = compressed_array[(sample_indices,)]
            #--- End: for
        #--- End: for

        return self.get_subspace(uarray, indices, copy=True)
    #--- End: def
    
#--- End: class



from builtins import object


class NetCDF(object):
    '''Mixin class for storing simple netCDF elements.

.. versionadded:: 1.7.0

    '''
    def _initialise_netcdf(self, source=None):
        '''Call this from inside the __init__ method of a class that inherits
from this mixin class.

:Parameters:

    source: optional

:Returns:

    `None`

        '''
        if source is None:
             netcdf = {}
        else:        
            try:
                netcdf = source._get_component('netcdf', None).copy()
            except AttributeError:
                netcdf = {}
        #--- End: if
        
        self._set_component('netcdf', netcdf, copy=False)
    #--- End: def

#--- End: class


class NetCDFDimension(NetCDF):
    '''Mixin class for accessing the netCDF dimension name.

.. versionadded:: 1.7.0

    '''
    def nc_del_dimension(self, *default):
        '''Remove the netCDF dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_get_dimension`, `nc_has_dimension`,
             `nc_set_dimension`

:Parameters:

    default: optional
        Return *default* if the netCDF dimension name has not been
        set.

:Returns:

    `str`
        The removed netCDF dimension name. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.nc_set_dimension('time')
>>> f.nc_has_dimension()
True
>>> f.nc_get_dimension()
'time'
>>> f.nc_del_dimension()
'time'
>>> f.nc_has_dimension()
False
>>> print(f.nc_get_dimension(None))
None
>>> print(f.nc_del_dimension(None))
None

        '''
        try:
            return self._get_component('netcdf').pop('dimension', *default)
        except KeyError:
            raise AttributeError(
                "{!r} object has no netCDF dimension name".format(
                    self.__class__.__name__))
    #--- End: def

    def nc_get_dimension(self, *default):
        '''Return the netCDF dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_dimension`, `nc_has_dimension`,
             `nc_set_dimension`

:Parameters:

    default: optional
        Return *default* if the netCDF dimension name has not been
        set.

:Returns:

    `str`
        The netCDF dimension name. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.nc_set_dimension('time')
>>> f.nc_has_dimension()
True
>>> f.nc_get_dimension()
'time'
>>> f.nc_del_dimension()
'time'
>>> f.nc_has_dimension()
False
>>> print(f.nc_get_dimension(None))
None
>>> print(f.nc_del_dimension(None))
None

        '''
        try:
            return self._get_component('netcdf')['dimension']
        except KeyError:
            if default:
                return default[0]

            raise AttributeError("{!r} has no netCDF dimension name".format(
                self.__class__.__name__))
    #--- End: def

    def nc_has_dimension(self):
        '''Whether the netCDF dimension name has been set.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_dimension`, `nc_get_dimension`,
             `nc_set_dimension`

:Returns:

    `bool`
        True if the netCDF dimension name has been set, otherwise
        False.

**Examples:**

>>> f.nc_set_dimension('time')
>>> f.nc_has_dimension()
True
>>> f.nc_get_dimension()
'time'
>>> f.nc_del_dimension()
'time'
>>> f.nc_has_dimension()
False
>>> print(f.nc_get_dimension(None))
None
>>> print(f.nc_del_dimension(None))
None

        '''
        return 'dimension' in self._get_component('netcdf')
    #--- End: def

    def nc_set_dimension(self, value):
        '''Set the netCDF dimension name.
        
.. versionadded:: 1.7.0

.. seealso:: `nc_del_dimension`, `nc_get_dimension`,
             `nc_has_dimension`

:Parameters:

    value: `str`
        The value for the netCDF dimension name.

:Returns:

    `None`

**Examples:**

>>> f.nc_set_dimension('time')
>>> f.nc_has_dimension()
True
>>> f.nc_get_dimension()
'time'
>>> f.nc_del_dimension()
'time'
>>> f.nc_has_dimension()
False
>>> print(f.nc_get_dimension(None))
None
>>> print(f.nc_del_dimension(None))
None

        '''
        self._get_component('netcdf')['dimension'] = value
    #--- End: def

#--- End: class


class NetCDFVariable(NetCDF):
    '''Mixin class for accessing the netCDF variable name.

.. versionadded:: 1.7.0

    '''
    def nc_del_variable(self, *default):
        '''Remove the netCDF variable name.

.. versionadded:: 1.7.0

.. seealso:: `nc_get_variable`, `nc_has_variable`, `nc_set_variable`

:Parameters:

    default: optional
        Return *default* if the netCDF variable name has not been set.

:Returns:

    `str`
        The removed netCDF variable name. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.nc_set_variable('tas')
>>> f.nc_has_variable()
True
>>> f.nc_get_variable()
'tas'
>>> f.nc_del_variable()
'tas'
>>> f.nc_has_variable()
False
>>> print(f.nc_get_variable(None))
None
>>> print(f.nc_del_variable(None))
None

        '''        
        try:
            return self._get_component('netcdf').pop('variable', *default)
        except KeyError:
            raise AttributeError(
                "{!r} object has no netCDF variable name".format(
                    self.__class__.__name__))
    #--- End: def
        
    def nc_get_variable(self, *default):
        '''Return the netCDF variable name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_variable`, `nc_has_variable`, `nc_set_variable`

:Parameters:

    default: optional
        Return *default* if the netCDF variable name has not been set.

:Returns:

    `str`
        The netCDF variable name. If unset then *default* is returned,
        if provided.

**Examples:**

>>> f.nc_set_variable('tas')
>>> f.nc_has_variable()
True
>>> f.nc_get_variable()
'tas'
>>> f.nc_del_variable()
'tas'
>>> f.nc_has_variable()
False
>>> print(f.nc_get_variable(None))
None
>>> print(f.nc_del_variable(None))
None

        '''
        try:
            return self._get_component('netcdf')['variable']
        except KeyError:
            if default:
                return default[0]

            raise AttributeError("{!r} has no netCDF variable name".format(
                self.__class__.__name__))
    #--- End: def

    def nc_has_variable(self):
        '''Whether the netCDF variable name has been set.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_variable`, `nc_get_variable`, `nc_set_variable`

:Returns:

    `bool`
        True if the netCDF variable name has been set, otherwise
        False.

**Examples:**

>>> f.nc_set_variable('tas')
>>> f.nc_has_variable()
True
>>> f.nc_get_variable()
'tas'
>>> f.nc_del_variable()
'tas'
>>> f.nc_has_variable()
False
>>> print(f.nc_get_variable(None))
None
>>> print(f.nc_del_variable(None))
None

        '''
        return 'variable' in self._get_component('netcdf')
    #--- End: def

    def nc_set_variable(self, value):
        '''Set the netCDF variable name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_variable`, `nc_get_variable`, `nc_has_variable`

:Parameters:

    value: `str`
        The value for the netCDF variable name.

:Returns:

    `None`

**Examples:**

>>> f.nc_set_variable('tas')
>>> f.nc_has_variable()
True
>>> f.nc_get_variable()
'tas'
>>> f.nc_del_variable()
'tas'
>>> f.nc_has_variable()
False
>>> print(f.nc_get_variable(None))
None
>>> print(f.nc_del_variable(None))
None

        '''
        self._get_component('netcdf')['variable'] = value
    #--- End: def

#--- End: class


class NetCDFSampleDimension(NetCDF):
    '''Mixin class for accessing the netCDF sample dimension name.

.. versionadded:: 1.7.0

    '''
    def nc_del_sample_dimension(self, *default):
        '''Remove the netCDF sample dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_get_sample_dimension`, `nc_has_sample_dimension`,
             `nc_set_sample_dimension`

:Parameters:

    default: optional        
        Return *default* if the netCDF sample dimension name has not
        been set.

:Returns:

    `str`
        The removed netCDF sample dimension name. If unset then
        *default* is returned, if provided.

**Examples:**

>>> f.nc_set_sample_dimension('time')
>>> f.nc_has_sample_dimension()
True
>>> f.nc_get_sample_dimension()
'time'
>>> f.nc_del_sample_dimension()
'time'
>>> f.nc_has_sample_dimension()
False
>>> print(f.nc_get_sample_dimension(None))
None
>>> print(f.nc_del_sample_dimension(None))
None

        '''        
        try:
            return self._get_component('netcdf').pop('sample_dimension', *default)
        except KeyError:
            raise AttributeError("{!r} has no netCDF sample dimension name".format(
                self.__class__.__name__))
    #--- End: def

    def nc_get_sample_dimension(self, *default):
        '''Return the netCDF sample dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_sample_dimension`, `nc_has_sample_dimension`,
             `nc_set_sample_dimension`

:Parameters:

    default: optional        
        Return *default* if the netCDF sample dimension name has not
        been set.

:Returns:

    `str`
        The netCDF sample dimension name. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.nc_set_sample_dimension('time')
>>> f.nc_has_sample_dimension()
True
>>> f.nc_get_sample_dimension()
'time'
>>> f.nc_del_sample_dimension()
'time'
>>> f.nc_has_sample_dimension()
False
>>> print(f.nc_get_sample_dimension(None))
None
>>> print(f.nc_del_sample_dimension(None))
None

        '''
        try:
            return self._get_component('netcdf')['sample_dimension']
        except KeyError:
            if default:
                return default[0]

            raise AttributeError("{!r} has no netCDF sample dimension name".format(
                self.__class__.__name__))
    #--- End: def

    def nc_has_sample_dimension(self):
        '''Whether the netCDF sample dimension name has been set.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_sample_dimension`, `nc_get_sample_dimension`,
             `nc_set_sample_dimension`

:Returns:
     
    `bool`
        True if the netCDF sample dimension name has been set,
        otherwise False.

**Examples:**

>>> f.nc_set_sample_dimension('time')
>>> f.nc_has_sample_dimension()
True
>>> f.nc_get_sample_dimension()
'time'
>>> f.nc_del_sample_dimension()
'time'
>>> f.nc_has_sample_dimension()
False
>>> print(f.nc_get_sample_dimension(None))
None
>>> print(f.nc_del_sample_dimension(None))
None

        '''
        return 'sample_dimension' in self._get_component('netcdf')
    #--- End: def

    def nc_set_sample_dimension(self, value):
        '''Set the netCDF sample dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_sample_dimension`, `nc_get_sample_dimension`,
             `nc_has_sample_dimension`
             
:Parameters:

    value: `str`
        The value for the netCDF sample dimension name.

:Returns:

    `None`

**Examples:**

>>> f.nc_set_sample_dimension('time')
>>> f.nc_has_sample_dimension()
True
>>> f.nc_get_sample_dimension()
'time'
>>> f.nc_del_sample_dimension()
'time'
>>> f.nc_has_sample_dimension()
False
>>> print(f.nc_get_sample_dimension(None))
None
>>> print(f.nc_del_sample_dimension(None))
None

        '''
        self._get_component('netcdf')['sample_dimension'] = value
    #--- End: def

#--- End: class

class NetCDFInstanceDimension(NetCDF):
    '''Mixin class for accessing the netCDF instance dimension name.

.. versionadded:: 1.7.0

    '''
    def nc_del_instance_dimension(self,*default):
        '''Return the netCDF instance dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_instance_dimension`, `nc_has_instance_dimension`,
             `nc_set_instance_dimension`

:Parameters:

    default: optional        
        Return *default* if the netCDF instance dimension name has not
        been set.

:Returns:

    `str`
        The netCDF instance dimension name. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.nc_set_instance_dimension('time')
>>> f.nc_has_instance_dimension()
True
>>> f.nc_get_instance_dimension()
'time'
>>> f.nc_del_instance_dimension()
'time'
>>> f.nc_has_instance_dimension()
False
>>> print(f.nc_get_instance_dimension(None))
None
>>> print(f.nc_del_instance_dimension(None))
None
 
        '''        
        try:
            return self._get_component('netcdf').pop('instance_dimension', *default)
        except KeyError:
            raise AttributeError("{!r} has no netCDF instance dimension name".format(
                self.__class__.__name__))
    #--- End: def

    def nc_get_instance_dimension(self, *default):
        '''Return the netCDF instance dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_instance_dimension`, `nc_has_instance_dimension`,
             `nc_set_instance_dimension`

:Parameters:

    default: optional        
        Return *default* if the netCDF instance dimension name has not
        been set.

:Returns:

    `str`
        The netCDF instance dimension name. If unset then *default* is
        returned, if provided.

**Examples:**

>>> f.nc_set_instance_dimension('time')
>>> f.nc_has_instance_dimension()
True
>>> f.nc_get_instance_dimension()
'time'
>>> f.nc_del_instance_dimension()
'time'
>>> f.nc_has_instance_dimension()
False
>>> print(f.nc_get_instance_dimension(None))
None
>>> print(f.nc_del_instance_dimension(None))
None
        '''
        try:
            return self._get_component('netcdf')['instance_dimension']
        except KeyError:
            if default:
                return default[0]

            raise AttributeError("{!r} has no netCDF instance dimension name".format(
                self.__class__.__name__))
    #--- End: def

    def nc_has_instance_dimension(self):
        '''Whether the netCDF instance dimension name has been set.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_instance_dimension`, `nc_get_instance_dimension`,
             `nc_set_instance_dimension`

:Returns:
     
    `bool`
        True if the netCDF instance dimension name has been set,
        otherwise False.

**Examples:**

>>> f.nc_set_instance_dimension('time')
>>> f.nc_has_instance_dimension()
True
>>> f.nc_get_instance_dimension()
'time'
>>> f.nc_del_instance_dimension()
'time'
>>> f.nc_has_instance_dimension()
False
>>> print(f.nc_get_instance_dimension(None))
None
>>> print(f.nc_del_instance_dimension(None))
None 
        '''
        return 'instance_dimension' in self._get_component('netcdf')
    #--- End: def

    def nc_set_instance_dimension(self, value):
        '''Set the netCDF instance dimension name.

.. versionadded:: 1.7.0

.. seealso:: `nc_del_instance_dimension`, `nc_get_instance_dimension`,
             `nc_has_instance_dimension`
             
:Parameters:

    value: `str`
        The value for the netCDF instance dimension name.

:Returns:

    `None`

**Examples:**

>>> f.nc_set_instance_dimension('time')
>>> f.nc_has_instance_dimension()
True
>>> f.nc_get_instance_dimension()
'time'
>>> f.nc_del_instance_dimension()
'time'
>>> f.nc_has_instance_dimension()
False
>>> print(f.nc_get_instance_dimension(None))
None
>>> print(f.nc_del_instance_dimension(None))
None
 
        '''
        self._get_component('netcdf')['instance_dimension'] = value
    #--- End: def

#--- End: class


class NetCDFDataVariable(NetCDF):
    '''Mixin class for accessing netCDF elements relating to a data
variable.

.. versionadded:: 1.7.0

    '''
    def nc_global_attributes(self, attributes=None):
        '''Return or replace the selection of properties to be written as
netCDF global attributes.

When multiple field constructs are being written to the same file, it
is not possible to create a netCDF global attribute from a property
that has different values for different fields being written. In this
case the property will not be written as a netCDF global attribute,
even if it has been selected as such by this method, but will appear
instead as an attribute on the netCDF data variable corresponding to
the field construct.

The `description of file contents
<http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#description-of-file-contents>`_
properties are always written as netCDF global attributes, if
possible, so selecting them is optional.

.. versionadded:: 1.7.0

.. seealso:: `cfdm.write`, `nc_unlimited_dimensions`

:Parameters:

    attributes: sequence of `str`, optional   
        Replace the exisiting selection of properties to be written
        out as netCDF global attributes.

        *Parameter example:*
          ``attributes=['project']``
        
        *Parameter example:*
          ``attributes=()``        

:Returns:

    `set`
        The selection of domain axis constructs prior to being
        changed, or the current selection if no changes were
        specified.

**Examples:**

>>> x = f.nc_global_attributes(['Conventions', 'project'])
>>> f.nc_global_attributes()
{'Conventions', 'project'}
>>> f.nc_global_attributes([])
{'Conventions', 'project'}
>>> f.nc_global_attributes()
set()

        '''
        out = self._get_component('netcdf').get('global_attributes')
        
        if out is None:
            out = set()

        if attributes:
            self._get_component('netcdf')['global_attributes'] = tuple(attributes)

        return set(out)
    #--- End: def
    
    def nc_unlimited_dimensions(self, axes=None):
        '''Return or replace the selection of domain axis constructs to be
written as netCDF unlimited dimensions.

By default output netCDF dimensions are not unlimited.

.. versionadded:: 1.7.0

.. seealso:: `cfdm.write`, `nc_global_attributes`

:Parameters:

    axes: sequence of `str`, optional   
        Replace the exisiting selection of domain axis constructs to
        be written out as netCDF unlimited dimensions. Domain axis
        constructs are identified by their construct identifiers.

        *Parameter example:*
          ``axes=['domainaxis0', 'domainaxis1']``
        
        *Parameter example:*
          ``axes=()``        

:Returns:

    `set`
        The selection of domain axis construct identifiers prior to
        being changed, or the current selection if no changes were
        specified.

**Examples:**

>>> x = f.nc_unlimited_dimensions(['domainaxis1'])
>>> f.nc_unlimited_dimensions()
{'domainaxis1'}
>>> f.nc_unlimited_dimensions([])
{'domainaxis1'}
>>> f.nc_unlimited_dimensions()
set()

        '''
        out = self._get_component('netcdf').get('unlimited_dimensions')

        if out is None:
            out = set()

        if axes:
            self._get_component('netcdf')['unlimited_dimensions'] = tuple(axes)

        return set(out)
#--- End: def


class NetCDFExternal(NetCDF):
    '''Mixin class for accessing the netCDF external variable status.

.. versionadded:: 1.7.0

    '''
    def nc_external(self, *external):
        '''Whether the construct correponds to an external netCDF variable.

.. versionadded:: 1.7.0

:Parameters:

    external: `bool`, optional
        Set the external status. If True then the construct will be
        written to a netCDF extenal file, if specified during the
        write process.

:Returns:

    `bool`
        The external status or, if the *external* parameter was set,
        the original status.

**Examples:**

>>> c.nc_external()
False
>>> c.nc_external(True)
False
>>> c.nc_external()
True

        '''
        old =  self._get_component('netcdf').get('external', False)
        if external:
            self._get_component('netcdf')['external'] = bool(external[0])

        return old
    #--- End: def

#--- End: class



from __future__ import print_function
from builtins import super

from . import Container


class Parameters(Container):
    '''Mixin class for parameters.

.. versionadded:: 1.7.0

    '''

    def __bool__(self):
        '''Called by the `bool` built-in function.

x.__bool__() <==> bool(x)

.. versionadded:: 1.7.0

        '''
        return bool(self.parameters())
    #--- End: def
        
    def __nonzero__(self):
        '''Called by the `bool` built-in function.

x.__nonzero__() <==> bool(x)

.. versionadded:: 1.7.0

        '''
        return bool(self.parameters())
    #--- End: def
        
    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7.0

        '''
        return 'Parameters: {0}'.format(', '.join(sorted(self.parameters())))
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_type=False):
        '''Whether two instances are the same.

Equality is strict by default. This means that:

* the named parameters must be the same, with the same values and data
  types, and vector-valued parameters must also have same the size and
  be element-wise equal (see the *ignore_data_type* parameter).

Two real numbers ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

Any type of object may be tested but, in general, equality is only
possible with another object of the same type, or a subclass of
one. See the *ignore_type* parameter.

:Parameters:

    other:
        The object to compare for equality.

    atol: `float`, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.

    rtol: `float`, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another object of the same type, or a
        subclass of one. If *ignore_type* is True then equality is
        possible for any object with a compatible API.

    verbose: `bool`, optional
        If True then print a verbose highlighting where the two
        instances differ.

:Returns: 

    `bool`
       Whether the two instances are equal.

**Examples:**

>>> d.equals(d)
True
>>> d.equals(d.copy())
True
>>> d.equals(None)
False

        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp in (True, False):
            return pp
        
        other = pp
                
        # Check that the coordinate conversion parameter terms match
        parameters0 = self.parameters()
        parameters1 = other.parameters()
        if set(parameters0) != set(parameters1):
            if verbose:
                print(
"{0}: Different parameter-valued terms ({1} != {2})".format(
    self.__class__.__name__,
    set(parameters0), set(parameters1)))
            return False

        # Check that the parameter values are equal
        for term, value0 in parameters0.items():            
            value1 = parameters1[term]  

            if value0 is None and value1 is None:
                # Parameter values are both None
                continue

            if not self._equals(value0, value1, rtol=rtol, atol=atol,
                                verbose=verbose,
                                ignore_data_type=True, #ignore_data_type,
                                ignore_fill_value=ignore_fill_value,
                                ignore_type=ignore_type):
                if verbose:
                    print(
"{}: Unequal {!r} terms ({!r} != {!r})".format( 
    self.__class__.__name__, term, value0, value1))
                return False
        #--- End: for

        # Still here? Then the two parameter collections are equal
        return True
    #--- End: def

#--- End: class



from . import mixin
from . import core


class Datum(mixin.Parameters,
            core.Datum):
    '''A datum component of a coordinate reference construct of the CF
data model.

A datum is a complete or partial definition of the zeroes of the
dimension and auxiliary coordinate constructs which define a
coordinate system.

The datum may contain the definition of a geophysical surface which
corresponds to the zero of a vertical coordinate construct, and this
may be required for both horizontal and vertical coordinate systems.

Elements of the datum not specified may be implied by the properties
of the dimension and auxiliary coordinate constructs referenced by the
`CoordinateReference` instance that contains the datum.

.. versionadded:: 1.7.0

    '''

#--- End: class



from builtins import super

from . import mixin
from . import core


class InteriorRing(mixin.NetCDFVariable,
                   mixin.PropertiesData,
                   core.InteriorRing):
    '''An interior ring array with properties.

For polygon geometries, an individual geometry may define an "interior
ring", i.e. a hole that is to be omitted from the cell extent (as
would occur, for example, for a cell describing the land area of a
region containing a lake). In this case an interior ring array is
required that records whether each polygon is to be included or
excluded from the cell, supplied by an interior ring variable in
CF-netCDF. The interior ring array spans the same domain axes as its
coordinate array, with the addition of an extra ragged dimension that
indexes the geometries for each cell.

.. versionadded:: 1.8.0

    '''
    def __init__(self, properties={}, data=None, source=None,
                 copy=True, _use_data=True):
        '''TODO
        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, field=None, key=None, _title=None,
             _create_title=True, _prefix=None, _level=0,
             _omit_properties=None):
        '''TODO
        '''
        if _create_title and _title is None: 
            _title = 'Interior Ring: ' + self.name(default='')

        return super().dump(display=display, field=field, key=key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title)
    #--- End: def
    
#--- End: class



from builtins import super

from . import mixin
from . import core


class Index(mixin.NetCDFVariable,
            mixin.NetCDFDimension,
            mixin.NetCDFInstanceDimension,
            mixin.NetCDFSampleDimension,
            mixin.PropertiesData,
            core.abstract.PropertiesData):
    '''An index variable required to uncompress a ragged array.

A collection of features stored using an indexed ragged array combines
all features along a single dimension (the sample dimension) such that
the values of each feature in the collection are interleaved.

The information needed to uncompress the data is stored in an index
variable that specifies the feature that each element of the sample
dimension belongs to.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties={}, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'long_name': 'which station this obs is for'}``

        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.

        The data array may also be set after initialisation with the
        `set_data` method.

    source: optional
        Initialize the properties and data from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, field=None, key=None, _title=None,
             _create_title=True, _prefix='', _level=0,
             _omit_properties=None):
        '''A full description of the index variable.

Returns a description of all properties, including those of
components, and provides selected values of all data arrays.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Index: ' + self.name(default='')

        return super().dump(display=display, field=field, key=key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title)
    #--- End: def
    
#--- End: class



from __future__ import print_function
import datetime
import os
import sys
import unittest

import numpy

import cfdm

verbose  = False
warnings = False

class create_fieldTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    def test_create_field(self):

        # Dimension coordinates
        dim1 = cfdm.DimensionCoordinate(data=cfdm.Data(numpy.arange(10.)))
        dim1.set_property('standard_name', 'grid_latitude')
        dim1.set_property('units', 'degrees')

        data = numpy.arange(9.) + 20
        data[-1] = 34
        dim0 = cfdm.DimensionCoordinate(data=cfdm.Data(data))
        dim0.set_property('standard_name', 'grid_longitude')
        dim0.set_property('units', 'degrees')

        array = dim0.get_array()

        array = numpy.array([array-0.5, array+0.5]).transpose((1,0))
        array[-2, 1] = 30
        array[-1, :] = [30, 36]
        dim0.set_bounds(cfdm.Bounds(data=cfdm.Data(array)))
        
        dim2 = cfdm.DimensionCoordinate(data=cfdm.Data([1.5]),
                                        bounds=cfdm.Bounds(data=cfdm.Data([[1, 2.]])))
        dim2.set_property('standard_name'         , 'atmosphere_hybrid_height_coordinate')
        dim2.set_property('computed_standard_name', 'altitude')
                      
        # Auxiliary coordinates
        ak = cfdm.DomainAncillary(data=cfdm.Data([10.])) #, 'm'))
        ak.set_property('units', 'm')
        ak.id = 'atmosphere_hybrid_height_coordinate_ak'
        ak.set_bounds(cfdm.Bounds(data=cfdm.Data([[5, 15.]]))) # , ak.Units)
        
        bk = cfdm.DomainAncillary(data=cfdm.Data([20.]))
        bk.id = 'atmosphere_hybrid_height_coordinate_bk'
        bk.set_bounds(cfdm.Bounds(data=cfdm.Data([[14, 26.]])))
        
        aux2 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(numpy.arange(-45, 45, dtype='int32').reshape(10, 9)))
        aux2.set_property('units', 'degree_N')
        aux2.set_property('standard_name', 'latitude')
        
        aux3 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(numpy.arange(60, 150, dtype='int32').reshape(9, 10)))
        aux3.set_property('standard_name', 'longitude')
        aux3.set_property('units', 'degreeE')

        array = numpy.ma.array(['alpha','beta','gamma','delta','epsilon',
                                'zeta','eta','theta','iota','kappa'], dtype='S')
        array[0] = numpy.ma.masked
        aux4 = cfdm.AuxiliaryCoordinate(data=cfdm.Data(array))
        aux4.set_property('long_name', 'greek_letters')

        # Cell measures
        msr0 = cfdm.CellMeasure(
            data=cfdm.Data(1+numpy.arange(90.).reshape(9, 10)*1234))
        msr0.set_measure('area')
        msr0.set_property('units', 'km2')
        
        # Data          
        data = cfdm.Data(numpy.arange(90.).reshape(10, 9))

        properties = {'units': 'm s-1'}
        
        f = cfdm.Field(properties=properties)
        f.set_property('standard_name', 'eastward_wind')

        axisX = f.set_construct(cfdm.DomainAxis(9))
        axisY = f.set_construct(cfdm.DomainAxis(10))
        axisZ = f.set_construct(cfdm.DomainAxis(1))

        f.set_data(data, axes=[axisY, axisX])
        
        x = f.set_construct(dim0, axes=axisX)
        y = f.set_construct(dim1, axes=axisY)
        z = f.set_construct(dim2, axes=[axisZ])

        lat   = f.set_construct(aux2, axes=[axisY, axisX])
        lon   = f.set_construct(aux3, axes=[axisX, axisY])
        greek = f.set_construct(aux4, axes=[axisY])

        ak = f.set_construct(ak, axes=axisZ)
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                        'grid_north_pole_latitude': 38.0,
                        'grid_north_pole_longitude': 190.0})
        
        datum = cfdm.Datum(parameters={'earth_radius': 6371007})
        
        ref0 = cfdm.CoordinateReference(
            coordinate_conversion=coordinate_conversion,
            datum=datum,
            coordinates=[x, y, lat, lon]
        )

        f.set_construct(msr0, axes=[axisX, axisY])

        f.set_construct(ref0)

        orog = cfdm.DomainAncillary(data=f.get_data())
        orog.set_property('standard_name', 'surface_altitude')
        orog.set_property('units', 'm')
        orog = f.set_construct(orog, axes=[axisY, axisX])

        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={'standard_name': 'atmosphere_hybrid_height_coordinate',
                        'computed_standard_name': 'altitude'},
            domain_ancillaries={'orog': orog,
                                'a'   : ak,
                                'b'   : bk})
        
        ref1 = cfdm.CoordinateReference(
            coordinates=[z],
            datum=datum,
            coordinate_conversion=coordinate_conversion
        #            parameters={'standard_name': 'atmosphere_hybrid_height_coordinate',
#                        'computed_standard_name': 'altitude',
 #                       'earth_radius' : 6371007,},
 #           domain_ancillaries={'orog': orog,
 #                               'a'   : ak,
#                                'b'   : bk},
        )
        
        ref1 = f.set_construct(ref1)

        # Field ancillary variables
        g = f.copy()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = 'ancillaryA'
        f.set_construct(anc, axes=[axisY, axisX])
        
        g = f[0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = 'ancillaryB'
        f.set_construct(anc, axes=axisX)

        g = f[..., 0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = 'ancillaryC'
        f.set_construct(anc, axes=[axisY])

        
        f.set_property('flag_values', numpy.array([1, 2, 4], 'int32'))
        f.set_property('flag_meanings', 'a bb ccc')
        f.set_property('flag_masks', [2, 1, 0])

        cm0 =  cfdm.CellMethod(axes=axisX,
                               method='mean',
                               properties={'interval': [cfdm.Data(1, 'day')],
                                           'comment' : 'ok'})
    
        cm1 =  cfdm.CellMethod(axes=[axisY],
                               method='maximum',
                               properties={'where' : 'sea'})

        f.set_construct(cm0)
        f.set_construct(cm1)

        if verbose:
            print(repr(f))
            print(f)
            print(f.constructs())
            print(f.construct_data_axes())
        
            f.dump()

        self.assertTrue(f.equals(f, verbose=True),
                        "Field f not equal to itself")

        self.assertTrue(f.equals(f.copy(), verbose=True),
                        "Field f not equal to a copy of itself")

        if verbose:
            print("####################################################")
            
        cfdm.write(f, self.filename, fmt='NETCDF3_CLASSIC', verbose=verbose)
#        sys.exit(0)

        g = cfdm.read(self.filename, verbose=verbose) #, squeeze=True)
#        for x in g:
#            x.print_read_report()

        if verbose:
            print(g)
            g[0].dump()

        array = g[0].get_construct(name='long_name:greek_letters').get_array()
        self.assertTrue(array[1] == b'beta', 'greek_letters = {!r}'.format(array))

        self.assertTrue(len(g) == 1, 'Read produced the wrong number of fields: {} != 1'.format(len(g)))

        g = g[0].squeeze()
        
#        g.dump()
#        print g
        self.assertTrue(sorted(f.constructs()) == sorted(g.constructs()),
                        '\n\nf (created in memory)\n{}\n\n{}\n\ng (read from disk)\n{}\n\n{}'.format(
                            sorted(f.constructs()),
                            sorted(f.constructs().items()),
                            sorted(g.constructs()),
                            sorted(g.constructs().items())))

        self.assertTrue(f.equals(f, verbose=True),
                        "Field f not equal to itself after having been written to disk")

        self.assertTrue(f.equals(f.copy(), verbose=True),
                        "Field f not equal to a copy of itself after having been written to disk")

        self.assertTrue(g.equals(g.copy(), verbose=True),
                        "Field g not equal to a copy of itself")

        if verbose:                    
#        print'f'
#        print f
#        print 'g'
#        print g
#        f.dump()
#        g.dump()
            print('g')
            g.dump()
            print('f')
            f.dump()
            
        self.assertTrue(g.equals(f, verbose=True),
                        "Field (f) not equal to itself read back in (g)")

        
        x = g.dump(display=False)
        x = f.dump(display=False)

        g = cfdm.read(self.filename, verbose=verbose,
                      extra='domain_ancillary', warnings=warnings)

        if verbose:            
            for x in g:
                x.print_read_report()

            print(g)
            
#        for x in g:
#            x.dump()
#        h = g.field('domainancillary2')
#        h.dump()
#        print h
#        
#
#        h = g.field('domainancillary1')
#        print h
#        
#        h = g.field('domainancillary0')
#        print h
#        
#        h = g.field('cellmeasure0')
#        print h

        
    #--- End: def

#--- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    unittest.main(verbosity=2)



from distutils.core import setup
import os
import fnmatch
import sys
import re

def find_package_data_files(directory):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, '*'):
                filename = os.path.join(root, basename)
                yield filename.replace('cfdm/', '', 1)


def _read(fname):
    """Returns content of a file.

    """
    fpath = os.path.dirname(__file__)
    fpath = os.path.join(fpath, fname)
    with open(fpath, 'r') as file_:
        return file_.read()

def _get_version():
    """Returns library version by inspecting __init__.py file.

    """
    return re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                     _read("cfdm/__init__.py"),
                     re.MULTILINE).group(1)



version      = _get_version()
packages     = ['cfdm']
#etc_files    = [f for f in find_package_data_files('cfdm/etc')]

long_description = """The **cfdm** Python package is a complete implementation of the `CF
data model <https://www.geosci-model-dev.net/10/4619/2017>`_, that
identifies the fundamental elements of the `CF conventions
<http://cfconventions.org/>`_ and shows how they relate to each other,
independently of the `netCDF
<https://www.unidata.ucar.edu/software/netcdf/>`_ encoding.

The central element defined by the CF data model is the **field
construct**, which corresponds to a CF-netCDF data variable with all
of its metadata.

The **cfdm** package can

* read field constructs from netCDF datasets,
* create new field constructs in memory,
* inspect field constructs,
* test whether two field constructs are the same,
* modify field construct metadata and data,
* create subspaces of field constructs,
* write field constructs to netCDF datasets on disk,
* incorporate, and create, metadata stored in external files, and
* read, write, and create data that have been compressed by convention
  (i.e. ragged or gathered arrays), whilst presenting a view of the
  data in its uncompressed form.

**Command line utility**

During installation the `cfdump` command line tool is also installed,
which generates text descriptions of the field constructs contained
in a netCDF dataset.

**Source code**

This project is hosted on a `GitHub repository
<https://github.com/NCAS-CMS/cfdm>`_ where you may access the most
up-to-date source."""

## --------------------------------------------------------------------
## Installation dependencies
## --------------------------------------------------------------------
#install_requires = [
#    'future>=0.16.0',
#    'netcdf4>=1.4.0',
#    'numpy==1.11',
#]

setup(name = "cfdm",
      long_description = long_description,
      version      = version,
      description  = "A complete implementation of the CF data model",
      author       = "David Hassell",
      maintainer   = "David Hassell",
      maintainer_email = "david.hassell@ncas.ac.uk",
      author_email = "david.hassell@ncas.ac.uk",
      url          = "https://ncas-cms.github.io/cfdm",
      download_url = "https://pypi.org/project/cfdm/#files",
      platforms    = ["Linux", "MacOS", "Windows"],
      keywords     = ['cf', 'netcdf', 'data', 'science',
                      'oceanography', 'meteorology', 'climate'],
      classifiers  = ["Development Status :: 4 - Beta",
                      "Intended Audience :: Science/Research", 
                      "License :: OSI Approved :: MIT License", 
                      "Topic :: Software Development",
                      "Topic :: Scientific/Engineering",
                      "Operating System :: OS Independent",
                      "Programming Language :: Python :: 2.7",
                      "Programming Language :: Python :: 3",
                      ],
      packages     = ['cfdm',
                      'cfdm.core',
                      'cfdm.core.abstract',
                      'cfdm.core.data',
                      'cfdm.core.data.abstract',
                      'cfdm.core.mixin',
                      'cfdm.data',
                      'cfdm.data.abstract',
                      'cfdm.data.mixin',
                      'cfdm.mixin',
                      'cfdm.read_write',
                      'cfdm.read_write.abstract',
                      'cfdm.read_write.netcdf',
                      'cfdm.test',],
      scripts      = ['scripts/cfdump'],
      install_requires = [
          'future>=0.16.0',
          'netcdf4>=1.4.0',
          'numpy>=1.11',
      ],
)
